from typing import Callable, Tuple

import torch

from tensordict.tensordict import TensorDict
from torch.distributions import Uniform

from rl4co.data.utils import save_tensordict_to_npz
from rl4co.envs.common.utils import Generator, get_sampler
from rl4co.utils.ops import get_distance
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def get_vehicle_capacity(num_loc: int) -> int:
    """Capacity should be 30 + num_loc/5 if num_loc > 20 as described in Liu et al. 2024 (POMO-MTL).
    For every N over 1000, we add 1 of capacity every 33.3 nodes to align with Ye et al. 2024 (GLOP),
    i.e. 260 at 2K nodes, 350 at 5K nodes and 500 at 10K nodes.
    Note that this serves as a demand scaler.
    """
    if num_loc > 1000:
        extra_cap = 1000 // 5 + (num_loc - 1000) // 33.3
    elif num_loc > 20:
        extra_cap = num_loc // 5
    else:
        extra_cap = 0
    return 30 + extra_cap


VARIANT_GENERATION_PRESETS = {
    "all": {"O": 0.5, "TW": 0.5, "L": 0.5, "B": 0.5},
    "single_feat": {"O": 0.5, "TW": 0.5, "L": 0.5, "B": 0.5},
    "single_feat_otw": {
        "O": 0.5,
        "TW": 0.5,
        "L": 0.5,
        "B": 0.5,
        "OTW": 0.5,
    },  # same training as Zhou et al. 2024
    "cvrp": {"O": 0.0, "TW": 0.0, "L": 0.0, "B": 0.0},
    "ovrp": {"O": 1.0, "TW": 0.0, "L": 0.0, "B": 0.0},
    "vrpb": {"O": 0.0, "TW": 0.0, "L": 0.0, "B": 1.0},
    "vrpl": {"O": 0.0, "TW": 0.0, "L": 1.0, "B": 0.0},
    "vrptw": {"O": 0.0, "TW": 1.0, "L": 0.0, "B": 0.0},
    "ovrptw": {"O": 1.0, "TW": 1.0, "L": 0.0, "B": 0.0},
    "ovrpb": {"O": 1.0, "TW": 0.0, "L": 0.0, "B": 1.0},
    "ovrpl": {"O": 1.0, "TW": 0.0, "L": 1.0, "B": 0.0},
    "vrpbl": {"O": 0.0, "TW": 0.0, "L": 1.0, "B": 1.0},
    "vrpbtw": {"O": 0.0, "TW": 1.0, "L": 0.0, "B": 1.0},
    "vrpltw": {"O": 0.0, "TW": 1.0, "L": 1.0, "B": 0.0},
    "ovrpbl": {"O": 1.0, "TW": 0.0, "L": 1.0, "B": 1.0},
    "ovrpbtw": {"O": 1.0, "TW": 1.0, "L": 0.0, "B": 1.0},
    "ovrpltw": {"O": 1.0, "TW": 1.0, "L": 1.0, "B": 0.0},
    "vrpbltw": {"O": 0.0, "TW": 1.0, "L": 1.0, "B": 1.0},
    "ovrpbltw": {"O": 1.0, "TW": 1.0, "L": 1.0, "B": 1.0},
}


class MTVRPGenerator(Generator):
    """MTVRP Generator.
    Class to generate instances of the MTVRP problem.
    If a variant is declared and Subsample is True, the generator will sample the problem based on the variant probabilities.
    By default, we use Mixed-Batch Training as in Berto et al. 2024 (RouteFinder), i.e. one batch can contain multiple variants.

    Example presets:
    - "all": Sample uniformly from 16 variants
    - "single_feat": Sample uniformly between CVRP, OVRP, VRPB, VRPL, VRPTW (as done in Liu et al. 2024 (MTPOMO))
    - "single_feat_otw": Sample uniformly between CVRP, OVRP, VRPB, VRPL, VRPTW, OVRPTW (as done in Zhou et al. 2024 (MVMoE))
    - "cvrp": Only CVRP (similarly for other variants)

    Args:
        num_loc: Number of locations to generate
        min_loc: Minimum location value
        max_loc: Maximum location value
        loc_distribution: Distribution to sample locations from
        capacity: Vehicle capacity. If None, get value based on `get_vehicle_capacity`
        min_demand: Minimum demand value
        max_demand: Maximum demand value
        min_backhaul: Minimum backhaul value
        max_backhaul: Maximum backhaul value
        scale_demand: Scale demand values (by default, generate between 1 and 10)
        max_time: Maximum time window value (at depot)
        backhaul_ratio: Fraction of backhauls (e.g. 0.2 means 20% of nodes are backhaul)
        distance_limit: Distance limit
        speed: Speed of vehicle. Defaults to 1
        subsample: If False, we always sample all attributes (i.e., OVRPBLTW)
            If true, we use the
        **kwargs: Additional keyword arguments
    """

    def __init__(
        self,
        num_loc: int = 20,
        min_loc: float = 0.0,
        max_loc: float = 1.0,
        loc_distribution: int | float | str | type | Callable = Uniform,
        capacity: float = None,
        min_demand: int = 1,
        max_demand: int = 10,
        min_backhaul: int = 1,
        max_backhaul: int = 10,
        scale_demand: bool = True,
        max_time: float = 4.6,
        backhaul_ratio: float = 0.2,
        distance_limit: float = 3.0,
        speed: float = 1.0,
        prob_open: float = 0.5,
        prob_time_window: float = 0.5,
        prob_limit: float = 0.5,
        prob_backhaul: float = 0.5,
        variant_preset=None,
        use_combinations=True,
        subsample=True,
        **kwargs,
    ) -> None:
        # Location distribution
        self.num_loc = num_loc
        self.min_loc = min_loc
        self.max_loc = max_loc
        if kwargs.get("loc_sampler", None) is not None:
            self.loc_sampler = kwargs["loc_sampler"]
        else:
            self.loc_sampler = get_sampler(
                "loc", loc_distribution, min_loc, max_loc, **kwargs
            )

        if capacity is None:
            capacity = get_vehicle_capacity(num_loc)
        self.capacity = capacity
        self.min_demand = min_demand
        self.max_demand = max_demand
        self.min_backhaul = min_backhaul
        self.max_backhaul = max_backhaul
        self.scale_demand = scale_demand
        self.backhaul_ratio = backhaul_ratio

        self.max_time = max_time
        self.distance_limit = distance_limit
        self.speed = speed

        assert not (
            subsample and (variant_preset is None)
        ), "Cannot use subsample if variant_preset is not specified. "
        if variant_preset is not None:
            log.info(f"Using variant generation preset {variant_preset}")
            variant_probs = VARIANT_GENERATION_PRESETS.get(variant_preset)
            assert (
                variant_probs is not None
            ), f"Variant generation preset {variant_preset} not found. \
                Available presets are {VARIANT_GENERATION_PRESETS.keys()} with probabilities {VARIANT_GENERATION_PRESETS.values()}"
        else:
            variant_probs = {
                "O": prob_open,
                "TW": prob_time_window,
                "L": prob_limit,
                "B": prob_backhaul,
            }
        # check probabilities
        for key, prob in variant_probs.items():
            assert 0 <= prob <= 1, f"Probability {key} must be between 0 and 1"
        self.variant_probs = variant_probs
        self.variant_preset = variant_preset
        if isinstance(variant_preset, str) and variant_preset != "all":
            log.warning(f"{variant_preset} selected. Will not use feature combination!")
            use_combinations = False
        self.use_combinations = use_combinations
        self.subsample = subsample

    def _generate(self, batch_size) -> TensorDict:
        # Locations
        locs = self.generate_locations(batch_size=batch_size, num_loc=self.num_loc)

        # Vehicle capacity (C, B) - applies to both linehaul and backhaul
        vehicle_capacity = torch.full(
            (*batch_size, 1), self.capacity, dtype=torch.float32
        )
        capacity_original = vehicle_capacity.clone()

        # linehaul demand / delivery (C) and backhaul / pickup demand (B)
        demand_linehaul, demand_backhaul = self.generate_demands(
            batch_size=batch_size, num_loc=self.num_loc
        )
        # add empty depot demands
        demand_linehaul = torch.cat(
            [torch.zeros(size=(*batch_size, 1)), demand_linehaul], dim=1
        )
        demand_backhaul = torch.cat(
            [torch.zeros(size=(*batch_size, 1)), demand_backhaul], dim=1
        )

        # Open (O)
        open_route = self.generate_open_route(shape=(*batch_size, 1))

        # Time windows (TW)
        speed = self.generate_speed(shape=(*batch_size, 1))
        time_windows, service_time = self.generate_time_windows(
            locs=locs,
            speed=speed,
        )

        # Distance limit (L)
        distance_limit = self.generate_distance_limit(shape=(*batch_size, 1), locs=locs)

        # scaling
        if self.scale_demand:
            demand_backhaul /= vehicle_capacity
            demand_linehaul /= vehicle_capacity
            vehicle_capacity /= vehicle_capacity

        # Put all variables together
        td = TensorDict(
            {
                "locs": locs,
                "demand_backhaul": demand_backhaul,  # (C)
                "demand_linehaul": demand_linehaul,  # (B)
                "distance_limit": distance_limit,  # (L)
                "time_windows": time_windows,  # (TW)
                "service_time": service_time,  # (TW)
                "vehicle_capacity": vehicle_capacity,  # (C)
                "capacity_original": capacity_original,  # unscaled capacity (C)
                "open_route": open_route,  # (O)
                "speed": speed,  # common
            },
            batch_size=batch_size,
        )

        if self.subsample:
            # Subsample problems based on given instructions
            return self.subsample_problems(td)
        else:
            # Not subsampling problems, i.e. return tensordict with all attributes
            return td

    def subsample_problems(self, td):
        """Create subproblems starting from seed probabilities depending on their variant.
        If random seed sampled in [0, 1] in batch is greater than prob, remove the constraint
        thus, if prob high, it is less likely to remove the constraint (i.e. prob=0.9, 90% chance to keep constraint)
        """
        batch_size = td.batch_size[0]

        variant_probs = torch.tensor(list(self.variant_probs.values()))

        if self.use_combinations:
            # in a batch, multiple variants combinations can be picked
            keep_mask = torch.rand(batch_size, 4) >= variant_probs  # O, TW, L, B
        else:
            # in a batch, only a variant can be picked.
            # we assign a 0.5 prob to the last variant (which is normal cvrp)
            if self.variant_preset in list(
                VARIANT_GENERATION_PRESETS.keys()
            ) and self.variant_preset not in (
                "all",
                "cvrp",
                "single_feat",
                "single_feat_otw",
            ):
                cvrp_prob = 0
            else:
                cvrp_prob = 0.5
            if self.variant_preset in ("all", "cvrp", "single_feat", "single_feat_otw"):
                indices = torch.distributions.Categorical(
                    torch.Tensor(list(self.variant_probs.values()) + [cvrp_prob])[
                        None
                    ].repeat(batch_size, 1)
                ).sample()
                if self.variant_preset == "single_feat_otw":
                    keep_mask = torch.zeros((batch_size, 6), dtype=torch.bool)
                    keep_mask[torch.arange(batch_size), indices] = True

                    # If keep_mask[:, 4] is True, make both keep_mask[:, 0] and keep_mask[:, 1] True
                    keep_mask[:, :2] |= keep_mask[:, 4:5]
                else:
                    keep_mask = torch.zeros((batch_size, 5), dtype=torch.bool)
                    keep_mask[torch.arange(batch_size), indices] = True
            else:
                # if the variant is specified, we keep the attributes with probability > 0
                keep_mask = torch.zeros((batch_size, 4), dtype=torch.bool)
                indices = torch.nonzero(variant_probs).squeeze()
                keep_mask[:, indices] = True

        td = self._default_open(td, ~keep_mask[:, 0])
        td = self._default_time_window(td, ~keep_mask[:, 1])
        td = self._default_distance_limit(td, ~keep_mask[:, 2])
        td = self._default_backhaul(td, ~keep_mask[:, 3])

        return td

    @staticmethod
    def _default_open(td, remove):
        td["open_route"][remove] = False
        return td

    @staticmethod
    def _default_time_window(td, remove):
        default_tw = torch.zeros_like(td["time_windows"])
        default_tw[..., 1] = float("inf")
        td["time_windows"][remove] = default_tw[remove]
        td["service_time"][remove] = torch.zeros_like(td["service_time"][remove])
        return td

    @staticmethod
    def _default_distance_limit(td, remove):
        td["distance_limit"][remove] = float("inf")
        return td

    @staticmethod
    def _default_backhaul(td, remove):
        # by default, where there is a backhaul, linehaul is 0. therefore, we add backhaul to linehaul
        # and set backhaul to 0 where we want to remove backhaul
        td["demand_linehaul"][remove] = (
            td["demand_linehaul"][remove] + td["demand_backhaul"][remove]
        )
        td["demand_backhaul"][remove] = 0
        return td

    def generate_locations(self, batch_size, num_loc) -> torch.Tensor:
        """Generate seed locations.

        Returns:
            locs: [B, N+1, 2] where the first location is the depot.
        """
        locs = torch.FloatTensor(*batch_size, num_loc + 1, 2).uniform_(
            self.min_loc, self.max_loc
        )
        return locs

    def generate_demands(self, batch_size: int, num_loc: int) -> torch.Tensor:
        """Classical lineahul demand / delivery from depot (C) and backhaul demand / pickup to depot (B) generation.
        Initialize the demand for nodes except the depot, which are added during reset.
        Demand sampling Following Kool et al. (2019), demands as integers between 1 and 10.
        Generates a slightly different distribution than using torch.randint.

        Returns:
            linehaul_demand: [B, N]
            backhaul_demand: [B, N]
        """
        linehaul_demand = (
            torch.FloatTensor(*batch_size, num_loc)
            .uniform_(self.min_demand - 1, self.max_demand - 1)
            .int()
            + 1
        ).float()
        # Backhaul demand sampling
        backhaul_demand = (
            torch.FloatTensor(*batch_size, num_loc)
            .uniform_(self.min_backhaul - 1, self.max_backhaul - 1)
            .int()
            + 1
        ).float()
        is_linehaul = torch.rand(*batch_size, num_loc) > self.backhaul_ratio
        backhaul_demand = (
            backhaul_demand * ~is_linehaul
        )  # keep only values where they are not linehauls
        linehaul_demand = linehaul_demand * is_linehaul
        return linehaul_demand, backhaul_demand

    def generate_time_windows(
        self,
        locs: torch.Tensor,
        speed: torch.Tensor,
    ) -> torch.Tensor:
        """Generate time windows (TW) and service times for each location including depot.
        We refer to the generation process in "Multi-Task Learning for Routing Problem with Cross-Problem Zero-Shot Generalization"
        (Liu et al., 2024). Note that another way to generate is from "Learning to Delegate for Large-scale Vehicle Routing" (Li et al, 2021) which
        is used in "MVMoE: Multi-Task Vehicle Routing Solver with Mixture-of-Experts" (Zhou et al, 2024). Note that however, in that case
        the distance limit would have no influence when time windows are present, since the tw for depot is the same as distance with speed=1.
        This function can be overridden for that implementation.
        See also https://github.com/RoyalSkye/Routing-MVMoE

        Args:
            locs: [B, N+1, 2] (depot, locs)
            speed: [B]

        Returns:
            time_windows: [B, N+1, 2]
            service_time: [B, N+1]
        """

        batch_size, n_loc = locs.shape[0], locs.shape[1] - 1  # no depot

        a, b, c = 0.15, 0.18, 0.2
        service_time = a + (b - a) * torch.rand(batch_size, n_loc)
        tw_length = b + (c - b) * torch.rand(batch_size, n_loc)
        d_0i = get_distance(locs[:, 0:1], locs[:, 1:])
        h_max = (self.max_time - service_time - tw_length) / d_0i * speed - 1
        tw_start = (1 + (h_max - 1) * torch.rand(batch_size, n_loc)) * d_0i / speed
        tw_end = tw_start + tw_length

        # Depot tw is 0, max_time
        time_windows = torch.stack(
            (
                torch.cat((torch.zeros(batch_size, 1), tw_start), -1),  # start
                torch.cat((torch.full((batch_size, 1), self.max_time), tw_end), -1),
            ),  # en
            dim=-1,
        )
        # depot service time is 0
        service_time = torch.cat((torch.zeros(batch_size, 1), service_time), dim=-1)
        return time_windows, service_time  # [B, N+1, 2], [B, N+1]

    def generate_distance_limit(
        self, shape: Tuple[int, int], locs: torch.Tensor
    ) -> torch.Tensor:
        """Generates distance limits (L) and checks their feasibilities.

        Returns:
            distance_limit: [B, 1]
        """
        # calculate distance of all locations to depot
        dist_to_depot = torch.cdist(locs, locs[:, 0:1, :], p=2)
        assert (
            dist_to_depot * 2 < self.distance_limit  # go back and forth
        ).all(), "Distance limit too low, not all nodes can be reached from the depot."
        return torch.full(shape, self.distance_limit, dtype=torch.float32)

    def generate_open_route(self, shape: Tuple[int, int]):
        """Generate open route flags (O). Here we could have a sampler but we simply return True here so all
        routes are open. Afterwards, we subsample the problems.
        """
        return torch.ones(shape, dtype=torch.bool)

    def generate_speed(self, shape: Tuple[int, int]):
        """We simply generate the speed as constant here"""
        # in this version, the speed is constant but this class may be overridden
        return torch.full(shape, self.speed, dtype=torch.float32)

    @staticmethod
    def save_data(td: TensorDict, path, compress: bool = False):
        save_tensordict_to_npz(td, path)

    @staticmethod
    def print_presets():
        for key, value in VARIANT_GENERATION_PRESETS.items():
            print(f"{key}: {value}")

    @staticmethod
    def available_variants(*args, **kwargs):
        # remove 'all', 'single_feat' from the list
        return list(VARIANT_GENERATION_PRESETS.keys())[3:]



from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl import REINFORCE
from rl4co.models.rl.reinforce.baselines import REINFORCEBaseline
from rl4co.models.zoo.am.policy import AttentionModelPolicy


class AttentionModel(REINFORCE):
    """Attention Model based on REINFORCE: https://arxiv.org/abs/1803.08475.
    Check :class:`REINFORCE` and :class:`rl4co.models.RL4COLitModule` for more details such as additional parameters  including batch size.

    Args:
        env: Environment to use for the algorithm
        policy: Policy to use for the algorithm
        baseline: REINFORCE baseline. Defaults to rollout (1 epoch of exponential, then greedy rollout baseline)
        policy_kwargs: Keyword arguments for policy
        baseline_kwargs: Keyword arguments for baseline
        **kwargs: Keyword arguments passed to the superclass
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: AttentionModelPolicy = None,
        baseline: REINFORCEBaseline | str = "rollout",
        policy_kwargs={},
        baseline_kwargs={},
        **kwargs,
    ):
        if policy is None:
            policy = AttentionModelPolicy(env_name=env.name, **policy_kwargs)

        super().__init__(env, policy, baseline, baseline_kwargs, **kwargs)



import os
import zipfile
from typing import Union, Callable

import torch
import numpy as np

from robust_downloader import download
from torch.distributions import Uniform
from tensordict.tensordict import TensorDict

from rl4co.data.utils import load_npz_to_tensordict
from rl4co.utils.pylogger import get_pylogger
from rl4co.envs.common.utils import get_sampler, Generator

log = get_pylogger(__name__)


class SMTWTPGenerator(Generator):
    """Data generator for the Single Machine Total Weighted Tardiness Problem (SMTWTP) environment
    
    Args:
        num_job: number of jobs
        min_time_span: lower bound of jobs' due time. By default, jobs' due time is uniformly sampled from (min_time_span, max_time_span)
        max_time_span: upper bound of jobs' due time. By default, it will be set to num_job / 2
        min_job_weight: lower bound of jobs' weights. By default, jobs' weights are uniformly sampled from (min_job_weight, max_job_weight)
        max_job_weight: upper bound of jobs' weights
        min_process_time: lower bound of jobs' process time. By default, jobs' process time is uniformly sampled from (min_process_time, max_process_time)
        max_process_time: upper bound of jobs' process time
    
    Returns:
        A TensorDict with the following key:
            job_due_time [batch_size, num_job + 1]: the due time of each job
            job_weight [batch_size, num_job + 1]: the weight of each job
            job_process_time [batch_size, num_job + 1]: the process time of each job
    """
    def __init__(
        self,
        num_job: int = 10,
        min_time_span: float = 0,
        max_time_span: float = None, # will be set to num_job / 2 by default. In DeepACO, it is set to num_job, which would be too simple
        min_job_weight: float = 0,
        max_job_weight: float = 1,
        min_process_time: float = 0,
        max_process_time: float = 1,
        **unused_kwargs
    ):
        self.num_job = num_job
        self.min_time_span = min_time_span
        self.max_time_span = num_job / 2 if max_time_span is None else max_time_span
        self.min_job_weight = min_job_weight
        self.max_job_weight = max_job_weight
        self.min_process_time = min_process_time
        self.max_process_time = max_process_time

        # SMTWTP environment doen't have any other kwargs
        if len(unused_kwargs) > 0:
            log.error(f"Found {len(unused_kwargs)} unused kwargs: {unused_kwargs}")

    def _generate(self, batch_size) -> TensorDict:
        batch_size = [batch_size] if isinstance(batch_size, int) else batch_size
        # Sampling according to Ye et al. (2023)
        job_due_time = (
            torch.FloatTensor(*batch_size, self.num_job + 1)
            .uniform_(self.min_time_span, self.max_time_span)
        )
        job_weight = (
            torch.FloatTensor(*batch_size, self.num_job + 1)
            .uniform_(self.min_job_weight, self.max_job_weight)
        )
        job_process_time = (
            torch.FloatTensor(*batch_size, self.num_job + 1)
            .uniform_(self.min_process_time, self.max_process_time)
        )

        # Rollouts begin at dummy node 0, whose features are set to 0
        job_due_time[:, 0] = 0
        job_weight[:, 0] = 0
        job_process_time[:, 0] = 0

        return TensorDict(
            {
                "job_due_time": job_due_time,
                "job_weight": job_weight,
                "job_process_time": job_process_time,
            },
            batch_size=batch_size,
        )



from rl4co.models.common.constructive.autoregressive.decoder import AutoregressiveDecoder
from rl4co.models.common.constructive.autoregressive.encoder import AutoregressiveEncoder
from rl4co.models.common.constructive.autoregressive.policy import AutoregressivePolicy



from rl4co.envs.routing.atsp.env import ATSPEnv
from rl4co.envs.routing.atsp.generator import ATSPGenerator
from rl4co.envs.routing.cvrp.env import CVRPEnv
from rl4co.envs.routing.cvrp.generator import CVRPGenerator
from rl4co.envs.routing.cvrpmvc.env import CVRPMVCEnv
from rl4co.envs.routing.cvrptw.env import CVRPTWEnv
from rl4co.envs.routing.cvrptw.generator import CVRPTWGenerator
from rl4co.envs.routing.mdcpdp.env import MDCPDPEnv
from rl4co.envs.routing.mdcpdp.generator import MDCPDPGenerator
from rl4co.envs.routing.mtsp.env import MTSPEnv
from rl4co.envs.routing.mtsp.generator import MTSPGenerator
from rl4co.envs.routing.mtvrp.env import MTVRPEnv
from rl4co.envs.routing.mtvrp.generator import MTVRPGenerator
from rl4co.envs.routing.op.env import OPEnv
from rl4co.envs.routing.op.generator import OPGenerator
from rl4co.envs.routing.pctsp.env import PCTSPEnv
from rl4co.envs.routing.pctsp.generator import PCTSPGenerator
from rl4co.envs.routing.pdp.env import PDPEnv, PDPRuinRepairEnv
from rl4co.envs.routing.pdp.generator import PDPGenerator
from rl4co.envs.routing.sdvrp.env import SDVRPEnv
from rl4co.envs.routing.shpp.env import SHPPEnv
from rl4co.envs.routing.shpp.generator import SHPPGenerator
from rl4co.envs.routing.spctsp.env import SPCTSPEnv
from rl4co.envs.routing.svrp.env import SVRPEnv
from rl4co.envs.routing.svrp.generator import SVRPGenerator
from rl4co.envs.routing.tsp.env import DenseRewardTSPEnv, TSPEnv, TSPkoptEnv
from rl4co.envs.routing.tsp.generator import TSPGenerator



import os
import platform
from ctypes import (
    Structure,
    CDLL,
    POINTER,
    c_int,
    c_double,
    c_char,
    sizeof,
    cast,
    byref,
)
from dataclasses import dataclass
from typing import List

import concurrent.futures
import numpy as np
import sys
import random
import time
import torch
from tensordict.tensordict import TensorDict

from rl4co.utils.ops import get_distance_matrix
from rl4co.utils.pylogger import get_pylogger


log = get_pylogger(__name__)


# Check if HGS-CVRP is installed
hgs_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "HGS-CVRP")
HGS_LIBRARY_FILEPATH = os.path.join(hgs_dir, "build", "libhgscvrp.so")
assert os.path.isfile(HGS_LIBRARY_FILEPATH)  # Failure case will be handled in the env.py


def local_search(
    td: TensorDict, actions: torch.Tensor, max_iterations: int = 1000, **kwargs
) -> torch.Tensor:
    """
    Improve the solution using local search for CVRP, based on PyVRP.

    Args:
        td: TensorDict, td from env with shape [batch_size,]
        actions: torch.Tensor, Tour indices with shape [batch_size, max_seq_len]
        max_trials: int, maximum number of trials for local search
        allow_infeasible_solution: bool, whether to allow infeasible solutions
        seed: int, random seed for local search
    Returns:
        torch.Tensor, Improved tour indices with shape [batch_size, max_seq_len]
    """

    # Convert tensors to numpy arrays
    # Note: to avoid the overhead of device transfer, we recommend to pass the tensors in cpu
    actions_np = actions.detach().cpu().numpy()  # [batch_size, max_seq_len]
    actions_np = np.pad(actions_np, ((0, 0), (1, 1)), mode="constant")
    positions_np = td["locs"].detach().cpu().numpy()  # [batch_size, num_loc + 1, 2]
    demands_np = td["demand"].detach().cpu().numpy()  # [batch_size, num_loc]
    demands_np = np.pad(demands_np, ((0, 0), (1, 0)), mode="constant")  # Add depot demand
    distances = td.get("distances", None)  # [batch_size, num_loc + 1, num_loc + 1]
    if distances is None:
        distances_np = get_distance_matrix(td["locs"]).numpy()
    else:
        distances_np = distances.detach().cpu().numpy()

    subroutes_all: List[List[List[int]]] = [get_subroutes(path) for path in actions_np]
    with concurrent.futures.ThreadPoolExecutor() as executor:
        futures = []
        for i in range(len(subroutes_all)):
            future = executor.submit(
                swapstar,
                demands_np[i],
                distances_np[i],
                positions_np[i],
                subroutes_all[i],
                count=max_iterations,
            )
            futures.append((i, future))

    worst_seq_len = positions_np.shape[1] * 2
    new_actions = np.zeros((actions_np.shape[0], worst_seq_len), dtype=np.int64)

    for i, future in futures:
        new_actions[i] = merge_subroutes(future.result(), worst_seq_len)

    # Remove heading and tailing zeros
    max_pos = np.max(np.where(new_actions != 0)[1])
    new_actions = new_actions[:, 1: max_pos + 1]
    new_actions = torch.from_numpy(new_actions).to(td.device)

    # Check the validity of the solution and use the original solution if the new solution is invalid
    isvalid = check_validity(td, new_actions)
    if not isvalid.all():    
        new_actions[~isvalid] = 0
        orig_valid_actions = actions[~isvalid]
        # pad if needed
        orig_max_pos = torch.max(torch.where(orig_valid_actions != 0)[1]) + 1
        if orig_max_pos > max_pos:
            new_actions = torch.nn.functional.pad(
                new_actions, (0, orig_max_pos - max_pos, 0, 0), mode="constant", value=0  # type: ignore
            )
        new_actions[~isvalid, :orig_max_pos] = orig_valid_actions[:, :orig_max_pos]
    return new_actions


def get_subroutes(path, end_with_zero = True) -> List[List[int]]:
    x = np.where(path == 0)[0]
    subroutes = []
    for i, j in zip(x, x[1:]):
        if j - i > 1:
            if end_with_zero:
                j = j + 1
            subroutes.append(path[i: j])
    return subroutes


def merge_subroutes(subroutes, length):
    route = np.zeros(length, dtype=np.int64)
    i = 0
    for r in subroutes:
        if len(r) > 2:
            r = r[:-1]  # remove the last zero
            route[i: i + len(r)] = r
            i += len(r)
    return route


def check_validity(td: TensorDict, actions: torch.Tensor) -> torch.Tensor:
    """
    Check the validity of the solution for CVRP and return a boolean tensor.
    Modified from CVRPEnv.check_solution_validity in rl4co/envs/routing/cvrp/env.py
    """
    # Check if tour is valid, i.e. contain 0 to n-1
    batch_size, graph_size = td["demand"].size()
    sorted_pi = actions.data.sort(1)[0]

    # Sorting it should give all zeros at front and then 1...n
    assert (
        torch.arange(1, graph_size + 1, out=sorted_pi.data.new())
        .view(1, -1)
        .expand(batch_size, graph_size)
        == sorted_pi[:, -graph_size:]
    ).all() and (sorted_pi[:, :-graph_size] == 0).all(), "Invalid tour"

    # Visiting depot resets capacity so we add demand = -capacity (we make sure it does not become negative)
    demand_with_depot = torch.cat((-td["vehicle_capacity"], td["demand"]), 1)
    d = demand_with_depot.gather(1, actions)

    used_cap = torch.zeros_like(td["demand"][:, 0])
    valid = torch.ones(batch_size, dtype=torch.bool)
    for i in range(actions.size(1)):
        used_cap += d[
            :, i
        ]  # This will reset/make capacity negative if i == 0, e.g. depot visited
        # Cannot use less than 0
        used_cap[used_cap < 0] = 0
        valid &= (
            used_cap <= td["vehicle_capacity"][:, 0] + 1e-5
        )
    return valid


########################## HGS-CVRP python wrapper ###########################
# Adapted from https://github.com/chkwon/PyHygese/blob/master/hygese/hygese.py


c_double_p = POINTER(c_double)
c_int_p = POINTER(c_int)
C_INT_MAX = 2 ** (sizeof(c_int) * 8 - 1) - 1
C_DBL_MAX = sys.float_info.max


def write_routes(routes: List[np.ndarray], filepath: str):
    with open(filepath, "w") as f:
        for i, r in enumerate(routes):
            f.write(f"Route #{i + 1}: "+' '.join([str(x) for x in r if x > 0])+"\n")
    return


def read_routes(filepath) -> List[np.ndarray]:
    routes = []
    with open(filepath, "r") as f:
        while 1:
            line = f.readline().strip()
            if line.startswith("Route"):
                routes.append(np.array([0, *map(int, line.split(":")[1].split()), 0]))
            else:
                break
    return routes


# Must match with AlgorithmParameters.h in HGS-CVRP: https://github.com/vidalt/HGS-CVRP
class CAlgorithmParameters(Structure):
    _fields_ = [
        ("nbGranular", c_int),
        ("mu", c_int),
        ("lambda", c_int),
        ("nbElite", c_int),
        ("nbClose", c_int),
        ("targetFeasible", c_double),
        ("seed", c_int),
        ("nbIter", c_int),
        ("timeLimit", c_double),
        ("useSwapStar", c_int),
    ]


@dataclass
class AlgorithmParameters:
    nbGranular: int = 20
    mu: int = 25
    lambda_: int = 40
    nbElite: int = 4
    nbClose: int = 5
    targetFeasible: float = 0.2
    seed: int = 0
    nbIter: int = 20000
    timeLimit: float = 0.0
    useSwapStar: bool = True

    @property
    def ctypes(self) -> CAlgorithmParameters:
        return CAlgorithmParameters(
            self.nbGranular,
            self.mu,
            self.lambda_,
            self.nbElite,
            self.nbClose,
            self.targetFeasible,
            self.seed,
            self.nbIter,
            self.timeLimit,
            int(self.useSwapStar),
        )


class _SolutionRoute(Structure):
    _fields_ = [("length", c_int), ("path", c_int_p)]


class _Solution(Structure):
    _fields_ = [
        ("cost", c_double),
        ("time", c_double),
        ("n_routes", c_int),
        ("routes", POINTER(_SolutionRoute)),
    ]


class RoutingSolution:
    def __init__(self, sol_ptr):
        if not sol_ptr:
            raise TypeError("The solution pointer is null.")

        self.cost = sol_ptr[0].cost
        self.time = sol_ptr[0].time
        self.n_routes = sol_ptr[0].n_routes
        self.routes = []
        for i in range(self.n_routes):
            r = sol_ptr[0].routes[i]
            path = r.path[0 : r.length]
            self.routes.append(path)


class Solver:
    def __init__(self, parameters=AlgorithmParameters(), verbose=False):
        if platform.system() == "Windows":
            hgs_library = CDLL(HGS_LIBRARY_FILEPATH, winmode=0)
        else:
            hgs_library = CDLL(HGS_LIBRARY_FILEPATH)

        self.algorithm_parameters = parameters
        self.verbose = verbose

        # solve_cvrp
        self._c_api_solve_cvrp = hgs_library.solve_cvrp
        self._c_api_solve_cvrp.argtypes = [
            c_int,
            c_double_p,
            c_double_p,
            c_double_p,
            c_double_p,
            c_double,
            c_double,
            c_char,
            c_char,
            c_int,
            POINTER(CAlgorithmParameters),
            c_char,
        ]
        self._c_api_solve_cvrp.restype = POINTER(_Solution)

        # solve_cvrp_dist_mtx
        self._c_api_local_search = hgs_library.local_search
        self._c_api_local_search.argtypes = [
            c_int,
            c_double_p,
            c_double_p,
            c_double_p,
            c_double_p,
            c_double_p,
            c_double,
            c_double,
            c_char,
            c_int,
            POINTER(CAlgorithmParameters),
            c_char,
            c_int,
            c_int,
        ]
        self._c_api_local_search.restype = c_int

        # delete_solution
        self._c_api_delete_sol = hgs_library.delete_solution
        self._c_api_delete_sol.restype = None
        self._c_api_delete_sol.argtypes = [POINTER(_Solution)]
    
    def local_search(self, data, routes: List[np.ndarray], count: int = 1) -> List[np.ndarray]:
        # required data
        demand = np.asarray(data["demands"])
        vehicle_capacity = data["vehicle_capacity"]
        n_nodes = len(demand)

        # optional depot
        depot = data.get("depot", 0)
        if depot != 0:
            raise ValueError("In HGS, the depot location must be 0.")

        # optional num_vehicles
        maximum_number_of_vehicles = data.get("num_vehicles", C_INT_MAX)

        # optional service_times
        service_times = data.get("service_times")
        if service_times is None:
            service_times = np.zeros(n_nodes)
        else:
            service_times = np.asarray(service_times)

        # optional duration_limit
        duration_limit = data.get("duration_limit")
        if duration_limit is None:
            is_duration_constraint = False
            duration_limit = C_DBL_MAX
        else:
            is_duration_constraint = True

        x_coords = data.get("x_coordinates")
        y_coords = data.get("y_coordinates")
        dist_mtx = data.get("distance_matrix")

        if x_coords is None or y_coords is None:
            assert dist_mtx is not None
            x_coords = np.zeros(n_nodes)
            y_coords = np.zeros(n_nodes)
        else:
            x_coords = np.asarray(x_coords)
            y_coords = np.asarray(y_coords)

        assert len(x_coords) == len(y_coords) == len(service_times) == len(demand)
        assert (x_coords >= 0.0).all()
        assert (y_coords >= 0.0).all()
        assert (service_times >= 0.0).all()
        assert (demand >= 0.0).all()

        dist_mtx = np.asarray(dist_mtx)
        assert dist_mtx.shape[0] == dist_mtx.shape[1]
        assert (dist_mtx >= 0.0).all()

        callid = (time.time_ns()*100000+random.randint(0,100000))%C_INT_MAX

        tmppath = "/tmp/route-{}".format(callid)
        resultpath = "/tmp/swapstar-result-{}".format(callid)
        write_routes(routes, tmppath)
        try:
            self._local_search(
                x_coords,
                y_coords,
                dist_mtx,
                service_times,
                demand,
                vehicle_capacity,
                duration_limit,
                is_duration_constraint,
                maximum_number_of_vehicles,
                self.algorithm_parameters,
                self.verbose,
                callid,
                count,
            )
            result = read_routes(resultpath)
        except Exception as e:
            print(e)
            result = routes
        else:
            os.remove(resultpath)
        finally:
            os.remove(tmppath)
        
        return result

    def _local_search(
        self,
        x_coords: np.ndarray,
        y_coords: np.ndarray,
        dist_mtx: np.ndarray,
        service_times: np.ndarray,
        demand: np.ndarray,
        vehicle_capacity: int,
        duration_limit: float,
        is_duration_constraint: bool,
        maximum_number_of_vehicles: int,
        algorithm_parameters: AlgorithmParameters,
        verbose: bool,
        callid: int,
        count:int,
    ):
        n_nodes = x_coords.size

        x_ct = x_coords.astype(c_double).ctypes
        y_ct = y_coords.astype(c_double).ctypes
        s_ct = service_times.astype(c_double).ctypes
        d_ct = demand.astype(c_double).ctypes

        m_ct = dist_mtx.reshape(n_nodes * n_nodes).astype(c_double).ctypes
        ap_ct = algorithm_parameters.ctypes


        # struct Solution *solve_cvrp_dist_mtx(
        # 	int n, double* x, double* y, double *dist_mtx, double *serv_time, double *dem,
        # 	double vehicleCapacity, double durationLimit, char isDurationConstraint,
        # 	int max_nbVeh, const struct AlgorithmParameters *ap, char verbose);
        sol_p = self._c_api_local_search(
            n_nodes,
            cast(x_ct, c_double_p),  # type: ignore
            cast(y_ct, c_double_p),  # type: ignore
            cast(m_ct, c_double_p),  # type: ignore
            cast(s_ct, c_double_p),  # type: ignore
            cast(d_ct, c_double_p),  # type: ignore
            vehicle_capacity,
            duration_limit,
            is_duration_constraint,
            maximum_number_of_vehicles,
            byref(ap_ct),
            verbose,
            callid,
            count,
        )

        result = sol_p
        return result


def swapstar(demands, matrix, positions, routes: List[np.ndarray], count=1):
    ap = AlgorithmParameters()
    hgs_solver = Solver(parameters=ap, verbose=False)

    data = dict()
    x = positions[:, 0]
    y = positions[:, 1]
    data['x_coordinates'] = x
    data['y_coordinates'] = y

    data['depot'] = 0
    data['demands'] = demands * 1000
    data["num_vehicles"] = len(routes)
    data['vehicle_capacity'] = 1000.001  # to avoid floating-point error

    # Solve with calculated distances
    data['distance_matrix'] = matrix
    result = hgs_solver.local_search(data, routes, count)
    return result



from dataclasses import dataclass
from typing import Optional

import numpy as np
import routefinder.baselines.pyvrp as pyvrp

from ortools.constraint_solver import pywrapcp, routing_enums_pb2
from tensordict import TensorDict
from torch import Tensor

from .constants import ORTOOLS_SCALING_FACTOR


def solve(instance: TensorDict, max_runtime: float, **kwargs) -> tuple[Tensor, Tensor]:
    """
    Solves an MTVRP instance with OR-Tools.

    Args:
        instance: The MTVRP instance to solve.
        max_runtime: The maximum runtime for the solver.

    Returns:
        A tuple containing the action and the cost, respectively.

    Note:
        This function depends on PyVRP's data converter to convert the MTVRP
        instance to an OR-Tools compatible format. Future versions should
        implement a direct conversion.
    """
    data = instance2data(instance)
    action, cost = _solve(data, max_runtime)
    cost /= ORTOOLS_SCALING_FACTOR
    cost *= -1

    return action, cost


@dataclass
class ORToolsData:
    """
    Convenient dataclass for instance data when using OR-Tools as solver.

    Args:
        depot: The depot index.
        distance_matrix: The distance matrix between locations.
        duration_matrix: The duration matrix between locations. This includes service times.
        num_vehicles: The number of vehicles.
        vehicle_capacities: The capacity of each vehicle.
        max_distance: The maximum distance a vehicle can travel.
        demands: The demands of each location.
        time_windows: The time windows for each location. Optional.
        backhauls: The pickup quantity for backhaul at each location.
    """

    depot: int
    distance_matrix: list[list[int]]
    duration_matrix: list[list[int]]
    num_vehicles: int
    vehicle_capacities: list[int]
    max_distance: int
    demands: list[int]
    time_windows: Optional[list[list[int]]]
    backhauls: Optional[list[int]]

    @property
    def num_locations(self) -> int:
        return len(self.distance_matrix)


def instance2data(instance: TensorDict) -> ORToolsData:
    """
    Converts an AnyVRP instance to an ORToolsData instance.
    """
    # TODO: Do not use PyVRP's data converter.
    data = pyvrp.instance2data(instance, ORTOOLS_SCALING_FACTOR)

    capacities = [
        veh_type.capacity
        for veh_type in data.vehicle_types()
        for _ in range(veh_type.num_available)
    ]
    max_distance = data.vehicle_type(0).max_distance

    demands = [0] + [client.delivery for client in data.clients()]
    backhauls = [0] + [client.pickup for client in data.clients()]
    service = [0] + [client.service_duration for client in data.clients()]

    tws = [[data.location(0).tw_early, data.location(0).tw_late]]
    tws += [[client.tw_early, client.tw_late] for client in data.clients()]

    # Set data to None if instance does not contain explicit values.
    default_tw = [0, np.iinfo(np.int64).max]
    if all(tw == default_tw for tw in tws):
        tws = None  # type: ignore

    if all(val == 0 for val in backhauls):
        backhauls = None  # type: ignore

    distances = data.distance_matrix().copy()
    durations = np.array(distances) + np.array(service)[:, np.newaxis]

    if backhauls is not None:
        # Serve linehauls before backhauls.
        linehaul = np.flatnonzero(np.array(demands) > 0)
        backhaul = np.flatnonzero(np.array(backhauls) > 0)
        distances[np.ix_(backhaul, linehaul)] = max_distance

    return ORToolsData(
        depot=0,
        distance_matrix=distances.tolist(),
        duration_matrix=durations.tolist(),
        num_vehicles=data.num_vehicles,
        vehicle_capacities=capacities,
        demands=demands,
        time_windows=tws,
        max_distance=max_distance,
        backhauls=backhauls,
    )


def _solve(data: ORToolsData, max_runtime: float, log: bool = False):
    """
    Solves an instance with OR-Tools.

    Args:
        data: The instance data.
        max_runtime: The maximum runtime in seconds.
        log: Whether to log the search.

    Returns:
        A tuple containing the action and the cost, respectively.
    """
    # Manager for converting between nodes (location indices) and index
    # (internal CP variable indices).
    manager = pywrapcp.RoutingIndexManager(
        data.num_locations, data.num_vehicles, data.depot
    )
    routing = pywrapcp.RoutingModel(manager)

    # Set arc costs equal to distances.
    distance_transit_idx = routing.RegisterTransitMatrix(data.distance_matrix)
    routing.SetArcCostEvaluatorOfAllVehicles(distance_transit_idx)

    # Max distance constraint.
    routing.AddDimension(
        distance_transit_idx,
        0,  # null distance slack
        data.max_distance,  # maximum distance per vehicle
        True,  # start cumul at zero
        "Distance",
    )

    # Vehicle capacity constraint.
    routing.AddDimensionWithVehicleCapacity(
        routing.RegisterUnaryTransitVector(data.demands),
        0,  # null capacity slack
        data.vehicle_capacities,  # vehicle maximum capacities
        True,  # start cumul to zero
        "Demand",
    )

    # Backhauls: this assumes that VRPB is implemented by forbidding arcs
    # that go from backhauls to linehauls.
    if data.backhauls is not None:
        routing.AddDimensionWithVehicleCapacity(
            routing.RegisterUnaryTransitVector(data.backhauls),
            0,  # null capacity slack
            data.vehicle_capacities,  # vehicle maximum capacities
            True,  # start cumul to zero
            "Backhaul",
        )

    # Time window constraints.
    if data.time_windows is not None:
        depot_tw_early = data.time_windows[data.depot][0]
        depot_tw_late = data.time_windows[data.depot][1]

        # The depot's late time window is a valid upper bound for the waiting
        # time and maximum duration per vehicle.
        routing.AddDimension(
            routing.RegisterTransitMatrix(data.duration_matrix),
            depot_tw_late,  # waiting time upper bound
            depot_tw_late,  # maximum duration per vehicle
            False,  # don't force start cumul to zero
            "Time",
        )
        time_dim = routing.GetDimensionOrDie("Time")

        for node, (tw_early, tw_late) in enumerate(data.time_windows):
            if node == data.depot:  # skip depot
                continue

            index = manager.NodeToIndex(node)
            time_dim.CumulVar(index).SetRange(tw_early, tw_late)

        # Add time window constraints for each vehicle start node.
        for node in range(data.num_vehicles):
            start = routing.Start(node)
            time_dim.CumulVar(start).SetRange(depot_tw_early, depot_tw_late)

        for node in range(data.num_vehicles):
            cumul_start = time_dim.CumulVar(routing.Start(node))
            routing.AddVariableMinimizedByFinalizer(cumul_start)

            cumul_end = time_dim.CumulVar(routing.End(node))
            routing.AddVariableMinimizedByFinalizer(cumul_end)

    # Setup search parameters.
    params = pywrapcp.DefaultRoutingSearchParameters()

    gls = routing_enums_pb2.LocalSearchMetaheuristic.GUIDED_LOCAL_SEARCH
    params.local_search_metaheuristic = gls

    params.time_limit.FromSeconds(int(max_runtime))  # only accepts int
    params.log_search = log

    solution = routing.SolveWithParameters(params)
    action = solution2action(data, manager, routing, solution)
    objective = solution.ObjectiveValue()

    return action, objective


def solution2action(data, manager, routing, solution) -> list[list[int]]:
    """
    Converts an OR-Tools solution to routes.
    """
    routes = []
    distance = 0  # for debugging

    for vehicle_idx in range(data.num_vehicles):
        index = routing.Start(vehicle_idx)
        route = []
        route_cost = 0

        while not routing.IsEnd(index):
            node = manager.IndexToNode(index)
            route.append(node)

            prev_index = index
            index = solution.Value(routing.NextVar(index))
            route_cost += routing.GetArcCostForVehicle(prev_index, index, vehicle_idx)

        if clients := route[1:]:  # ignore depot
            routes.append(clients)
            distance += route_cost

    return [visit for route in routes for visit in route + [0]]



import tensordict
import torch

from packaging import version
from tensordict.tensordict import TensorDict
from torch.utils.data import Dataset

# Checks were removed in tensordict 0.5.0, so we should not pass the kwargs
if version.parse(tensordict.__version__) <= version.parse("0.4.0"):
    td_kwargs = {"_run_checks": False}
else:
    td_kwargs = {}


class FastTdDataset(Dataset):
    """
    Note:
        Check out the issue on tensordict for more details:
        https://github.com/pytorch-labs/tensordict/issues/374.
    """

    def __init__(self, td: TensorDict):
        self.data_len = td.batch_size[0]
        self.data = td

    def __len__(self):
        return self.data_len

    def __getitems__(self, idx):
        return self.data[idx]

    def add_key(self, key, value):
        return ExtraKeyDataset(self, value, key_name=key)

    @staticmethod
    def collate_fn(batch: dict | TensorDict):
        """Collate function compatible with TensorDicts that reassembles a list of dicts."""
        return batch


class TensorDictDataset(Dataset):
    """Dataset compatible with TensorDicts with low CPU usage.
    Fast loading but somewhat slow instantiation due to list comprehension since we
    "disassemble" the TensorDict into a list of dicts.

    Note:
        Check out the issue on tensordict for more details:
        https://github.com/pytorch-labs/tensordict/issues/374.
    """

    def __init__(self, td: TensorDict):
        self.data_len = td.batch_size[0]
        self.data = [
            {key: value[i] for key, value in td.items()} for i in range(self.data_len)
        ]

    def __len__(self):
        return self.data_len

    def __getitem__(self, idx):
        return self.data[idx]

    def add_key(self, key, value):
        return ExtraKeyDataset(self, value, key_name=key)

    @staticmethod
    def collate_fn(batch: dict | TensorDict):
        """Collate function compatible with TensorDicts that reassembles a list of dicts."""
        return TensorDict(
            {key: torch.stack([b[key] for b in batch]) for key in batch[0].keys()},
            batch_size=torch.Size([len(batch)]),
            **td_kwargs,
        )


class ExtraKeyDataset(TensorDictDataset):
    """Dataset that includes an extra key to add to the data dict.
    This is useful for adding a REINFORCE baseline reward to the data dict.
    Note that this is faster to instantiate than using list comprehension.
    """

    def __init__(self, dataset: TensorDictDataset, extra: torch.Tensor, key_name="extra"):
        self.data_len = len(dataset)
        assert self.data_len == len(extra), "Data and extra must be same length"
        self.data = dataset.data
        self.extra = extra
        self.key_name = key_name

    def __getitem__(self, idx):
        data = self.data[idx]
        data[self.key_name] = self.extra[idx]
        return data


class TensorDictDatasetFastGeneration(Dataset):
    """Dataset compatible with TensorDicts.
    Similar performance in loading to list comprehension, but is faster in instantiation
    than :class:`TensorDictDatasetList` (more than 10x faster).

    Warning:
        Note that directly indexing TensorDicts may be faster in creating the dataset
        but uses > 3x more CPU. We may generally recommend using the :class:`TensorDictDatasetList`

    Note:
        Check out the issue on tensordict for more details:
        https://github.com/pytorch-labs/tensordict/issues/374.
    """

    def __init__(self, td: TensorDict):
        self.data = td

    def __len__(self):
        return len(self.data)

    def __getitems__(self, index):
        # Tricks:
        # - batched data loading with `__getitems__` for faster loading
        # - avoid directly indexing TensorDicts for faster loading
        return TensorDict(
            {key: item[index] for key, item in self.data.items()},
            batch_size=torch.Size([len(index)]),
            **td_kwargs,
        )

    def add_key(self, key, value):
        self.data.update({key: value})  # native method
        return self

    @staticmethod
    def collate_fn(batch: dict | TensorDict):
        """Equivalent to collating with `lambda x: x`"""
        return batch



from typing import Callable

import torch.nn as nn

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl.reinforce.baselines import REINFORCEBaseline
from rl4co.models.zoo.am import AttentionModel, AttentionModelPolicy
from rl4co.models.zoo.pomo import POMO
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class MVMoE_POMO(POMO):
    """MVMoE Model for neural combinatorial optimization based on POMO and REINFORCE
    Please refer to Zhou et al. (2024) <https://arxiv.org/abs/2405.01029>.
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: nn.Module = None,
        policy_kwargs={},
        baseline: str = "shared",
        num_augment: int = 8,
        augment_fn: str | Callable = "dihedral8",
        first_aug_identity: bool = True,
        feats: list = None,
        num_starts: int = None,
        moe_kwargs: dict = None,
        **kwargs,
    ):
        if moe_kwargs is None:
            moe_kwargs = {
                "encoder": {
                    "hidden_act": "ReLU",
                    "num_experts": 4,
                    "k": 2,
                    "noisy_gating": True,
                },
                "decoder": {
                    "light_version": True,
                    "num_experts": 4,
                    "k": 2,
                    "noisy_gating": True,
                },
            }

        if policy is None:
            policy_kwargs_ = {
                "num_encoder_layers": 6,
                "normalization": "instance",
                "use_graph_context": False,
                "moe_kwargs": moe_kwargs,
            }
            policy_kwargs.update(policy_kwargs_)
            policy = AttentionModelPolicy(env_name=env.name, **policy_kwargs)

        # Initialize with the shared baseline
        super(MVMoE_POMO, self).__init__(
            env,
            policy,
            policy_kwargs,
            baseline,
            num_augment,
            augment_fn,
            first_aug_identity,
            feats,
            num_starts,
            **kwargs,
        )


class MVMoE_AM(AttentionModel):
    """MVMoE Model for neural combinatorial optimization based on AM and REINFORCE
    Please refer to Zhou et al. (2024) <https://arxiv.org/abs/2405.01029>.
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: AttentionModelPolicy = None,
        baseline: REINFORCEBaseline | str = "rollout",
        policy_kwargs={},
        baseline_kwargs={},
        moe_kwargs: dict = None,
        **kwargs,
    ):
        if moe_kwargs is None:
            moe_kwargs = {
                "encoder": {
                    "hidden_act": "ReLU",
                    "num_experts": 4,
                    "k": 2,
                    "noisy_gating": True,
                },
                "decoder": {
                    "light_version": True,
                    "out_bias": False,
                    "num_experts": 4,
                    "k": 2,
                    "noisy_gating": True,
                },
            }

        if policy is None:
            policy_kwargs_ = {
                "moe_kwargs": moe_kwargs,
            }
            policy_kwargs.update(policy_kwargs_)
            policy = AttentionModelPolicy(env_name=env.name, **policy_kwargs)

        # Initialize with the shared baseline
        super(MVMoE_AM, self).__init__(
            env, policy, baseline, policy_kwargs, baseline_kwargs, **kwargs
        )



from typing import Optional

import torch

from tensordict.tensordict import TensorDict

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.utils.pylogger import get_pylogger

from .generator import MCPGenerator

log = get_pylogger(__name__)


class MCPEnv(RL4COEnvBase):
    """Maximum Coverage Problem (MCP) environment
    At each step, the agent chooses a set. The reward is 0 unless enough number of sets are chosen.
    The reward is the total weights of the covered items (i.e., items in any chosen set).

    Observations:
        - the weights of items
        - the membership of items in sets
        - the number of sets to choose

    Constraints:
        - the given number of sets must be chosen

    Finish condition:
        - the given number of sets are chosen

    Reward:
        - the total weights of the covered items (i.e., items in any chosen set)

    Args:
        generator: MCPGenerator instance as the data generator
        generator_params: parameters for the generator
    """

    name = "mcp"

    def __init__(
        self,
        generator: MCPGenerator = None,
        generator_params: dict = {},
        check_solution=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = MCPGenerator(**generator_params)
        self.generator = generator
        self.check_solution = check_solution
        self._make_spec(self.generator)

    def _step(self, td: TensorDict) -> TensorDict:
        # action: [batch_size, 1]; the set to be chosen in each instance
        batch_size = td["action"].shape[0]
        selected = td["action"]

        # Update set selection status
        chosen = td["chosen"].clone()  # (batch_size, n_sets)
        chosen[torch.arange(batch_size).to(td.device), selected] = True

        # We are done if we choose enough sets
        done = td["i"] >= (td["n_sets_to_choose"] - 1)

        # The reward is calculated outside via get_reward for efficiency, so we set it to -inf here
        reward = torch.ones_like(done) * float("-inf")

        remaining_sets = ~chosen  # (batch_size, n_sets)

        chosen_membership = chosen.unsqueeze(-1) * td["membership"]
        chosen_membership_nonzero = chosen_membership.nonzero()
        remaining_membership = remaining_sets.unsqueeze(-1) * td["membership"]

        batch_indices, set_indices, item_indices = chosen_membership_nonzero.T
        chosen_items_indices = chosen_membership[
            batch_indices, set_indices, item_indices
        ].long()

        batch_size, n_items = td["weights"].shape

        # We have batch_indices and chosen_items_indices
        # chosen_items: (batch_size, n_items)
        # for each i, chosen_items[batch_size[i], chosen_items_indices[i]] += 1
        chosen_items = torch.zeros(batch_size, n_items + 1, device=td.device)
        chosen_items[batch_indices, chosen_items_indices] += 1
        chosen_items = chosen_items[:, 1:]  # Remove the first column (invalid zeros)

        # chosen_item[i, j] > 0 means item j is chosen in batch i
        covered_items = (chosen_items > 0).float()  # (batch_size, n_items)
        remaining_items = 1.0 - covered_items  # (batch_size, n_items)

        # We cannot choose the already-chosen sets
        action_mask = ~chosen

        td.update(
            {
                "membership": remaining_membership,  # (batch_size, n_sets, max_size)
                "weights": td["weights"] * remaining_items,  # (batch_size, n_items)
                "chosen": chosen,
                "i": td["i"] + 1,
                "action_mask": action_mask,
                "reward": reward,
                "done": done,
            }
        )
        return td

    def _reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        self.to(td.device)

        return TensorDict(
            {
                # given information; constant for each given instance
                "orig_membership": td["membership"],  # (batch_size, n_sets, max_size)
                "membership": td["membership"],  # (batch_size, n_sets, max_size)
                "orig_weights": td["weights"],  # (batch_size, n_items)
                "weights": td["weights"],  # (batch_size, n_items)
                "n_sets_to_choose": td["n_sets_to_choose"],  # (batch_size, 1)
                # states changed by actions
                "chosen": torch.zeros(
                    *td["membership"].shape[:-1], dtype=torch.bool, device=td.device
                ),  # each entry is binary; 1 iff the corresponding set is chosen
                "i": torch.zeros(
                    *batch_size, dtype=torch.int64, device=td.device
                ),  # the number of sets we have chosen
                "action_mask": torch.ones(
                    *td["membership"].shape[:-1], dtype=torch.bool, device=td.device
                ),
            },
            batch_size=batch_size,
        )

    def _make_spec(self, generator: MCPGenerator):
        # TODO: make spec
        pass

    def _get_reward(self, td: TensorDict, actions: torch.Tensor) -> torch.Tensor:
        if self.check_solution:
            self.check_solution_validity(td, actions)

        membership = td[
            "orig_membership"
        ]  # (batch_size, n_sets, max_size); membership[i, j] = the items in set j in batch i (with 0 padding)
        weights = td["orig_weights"]  # (batch_size, n_items)
        chosen_sets = td["chosen"]  # (batch_size, n_set); 1 if chosen, 0 otherwise

        chosen_membership = chosen_sets.unsqueeze(-1) * membership
        chosen_membership_nonzero = chosen_membership.nonzero()

        batch_indices, set_indices, item_indices = chosen_membership_nonzero.T
        chosen_items_indices = chosen_membership[
            batch_indices, set_indices, item_indices
        ].long()

        batch_size, n_items = weights.shape

        # We have batch_indices and chosen_items_indices
        # chosen_items: (batch_size, n_items)
        # For each i, chosen_items[batch_size[i], chosen_items_indices[i]] += 1
        chosen_items = torch.zeros(batch_size, n_items + 1, device=td.device)
        chosen_items[batch_indices, chosen_items_indices] += 1
        chosen_items = chosen_items[:, 1:]  # remove the first column

        # chosen_item[i, j] > 0 means item j is chosen in batch i
        chosen_items = (chosen_items > 0).float()
        # Compute the total weights of chosen items
        chosen_weights = torch.sum(chosen_items * weights, dim=-1)

        return chosen_weights

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor) -> None:
        # TODO: check solution validity
        pass

    @staticmethod
    def local_search(td: TensorDict, actions: torch.Tensor, **kwargs) -> torch.Tensor:
        # TODO: local search
        pass

    @staticmethod
    def get_num_starts(td):
        return td["action_mask"].shape[-1]

    @staticmethod
    def select_start_nodes(td, num_starts):
        num_sets = td["action_mask"].shape[-1]
        return (
            torch.arange(num_starts, device=td.device).repeat_interleave(td.shape[0])
            % num_sets
        )



import matplotlib.pyplot as plt
import numpy as np
import torch

from rl4co.utils.ops import gather_by_index
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def render(td, actions=None, ax=None):
    if ax is None:
        # Create a plot of the nodes
        _, ax = plt.subplots()

    td = td.detach().cpu()

    if actions is None:
        actions = td.get("action", None)

    # If batch_size greater than 0 , we need to select the first batch element
    if td.batch_size != torch.Size([]):
        td = td[0]
        actions = actions[0]

    locs = td["locs"]

    # Gather locs in order of action if available
    if actions is None:
        log.warning("No action in TensorDict, rendering unsorted locs")
    else:
        actions = actions.detach().cpu()
        locs = gather_by_index(locs, actions, dim=0)

    # Cat the first node to the end to complete the tour
    locs = torch.cat((locs, locs[0:1]))
    x, y = locs[:, 0], locs[:, 1]

    # Plot the visited nodes
    ax.scatter(x, y, color="tab:blue")

    # Add arrows between visited nodes as a quiver plot
    dx, dy = np.diff(x), np.diff(y)
    ax.quiver(x[:-1], y[:-1], dx, dy, scale_units="xy", angles="xy", scale=1, color="k")

    return ax



from typing import Callable

import torch

from tensordict.tensordict import TensorDict
from torch.distributions import Uniform

from rl4co.envs.common.utils import Generator, get_sampler
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def remove_repeat(x: torch.Tensor) -> torch.Tensor:
    """
    Remove the repeated elements in each row (i.e., the last dimension) of the input tensor x,
    and change the repeated elements to 0

    Ref: https://stackoverflow.com/questions/62300404

    Args:
        x: input tensor
    """

    # sorting the rows so that duplicate values appear together
    # e.g., first row: [1, 2, 3, 3, 3, 4, 4]
    y, indices = x.sort(dim=-1)

    # subtracting, so duplicate values will become 0
    # e.g., first row: [1, 2, 3, 0, 0, 4, 0]
    y[..., 1:] *= ((y[..., 1:] - y[..., :-1]) != 0).long()

    # retrieving the original indices of elements
    indices = indices.sort(dim=-1)[1]

    # re-organizing the rows following original order
    # e.g., first row: [1, 2, 3, 4, 0, 0, 0]
    return torch.gather(y, -1, indices)


class MCPGenerator(Generator):
    """Data generator for the Maximum Coverage Problem (MCP).

    Args:
        num_items: number of items in the MCP
        num_sets: number of sets in the MCP
        min_weight: minimum value for the item weights
        max_weight: maximum value for the item weights
        min_size: minimum size for the sets
        max_size: maximum size for the sets
        n_sets_to_choose: number of sets to choose in the MCP

    Returns:
        A TensorDict with the following keys:
            membership [batch_size, num_sets, max_size]: membership of items in sets
            weights [batch_size, num_items]: weights of the items
            n_sets_to_choose [batch_size, 1]: number of sets to choose in the MCP
    """

    def __init__(
        self,
        num_items: int = 200,
        num_sets: int = 100,
        min_weight: int = 1,
        max_weight: int = 10,
        min_size: int = 5,
        max_size: int = 15,
        n_sets_to_choose: int = 10,
        size_distribution: int | float | str | type | Callable = Uniform,
        weight_distribution: int | float | str | type | Callable = Uniform,
        **kwargs,
    ):
        self.num_items = num_items
        self.num_sets = num_sets
        self.min_weight = min_weight
        self.max_weight = max_weight
        self.min_size = min_size
        self.max_size = max_size
        self.n_sets_to_choose = n_sets_to_choose

        # Set size distribution
        if kwargs.get("size_sampler", None) is not None:
            self.size_sampler = kwargs["size_sampler"]
        else:
            self.size_sampler = get_sampler(
                "size", size_distribution, min_size, max_size + 1, **kwargs
            )

        # Item weight distribution
        if kwargs.get("weight_sampler", None) is not None:
            self.weight_sampler = kwargs["weight_sampler"]
        else:
            self.weight_sampler = get_sampler(
                "weight", weight_distribution, min_weight, max_weight + 1, **kwargs
            )

    def _generate(self, batch_size) -> TensorDict:
        try:
            batch_size = batch_size[0]
        except TypeError:
            batch_size = batch_size

        # Sample item weights
        weights_tensor = self.weight_sampler.sample((batch_size, self.num_items))
        weights_tensor = torch.floor(weights_tensor)
        weights_tensor = torch.clamp(weights_tensor, self.min_weight, self.max_weight)

        # Sample set sizes
        set_sizes = self.size_sampler.sample((batch_size, self.num_sets))
        set_sizes = torch.floor(set_sizes).long()
        set_sizes = torch.clamp(set_sizes, self.min_size, self.max_size)
        max_size = set_sizes.max().item()

        # Create membership tensor
        membership_tensor_max_size = torch.randint(
            1, self.num_items + 1, (batch_size, self.num_sets, max_size)
        )

        cutoffs_masks = torch.arange(self.max_size).view(1, 1, -1) < set_sizes.unsqueeze(
            -1
        )
        # Take the masked elements, 0 means the item is invalid
        membership_tensor = (
            membership_tensor_max_size * cutoffs_masks
        )  # (batch_size, num_sets, max_size)

        # Remove repeated items in each set
        membership_tensor = remove_repeat(membership_tensor)

        return TensorDict(
            {
                "membership": membership_tensor.float(),  # (batch_size, num_sets, max_size)
                "weights": weights_tensor.float(),  # (batch_size, num_items)
                "n_sets_to_choose": torch.ones(batch_size, 1)
                * self.n_sets_to_choose,  # (batch_size, 1)
            },
            batch_size=batch_size,
        )



from typing import Optional

import torch

from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, Unbounded

from rl4co.data.utils import (
    load_npz_to_tensordict,
    load_solomon_instance,
    load_solomon_solution,
)
from rl4co.envs.routing.cvrp.env import CVRPEnv
from rl4co.utils.ops import gather_by_index, get_distance

from ..cvrp.generator import CVRPGenerator
from .generator import CVRPTWGenerator
from .render import render


class CVRPTWEnv(CVRPEnv):
    """Capacitated Vehicle Routing Problem with Time Windows (CVRPTW) environment.
    Inherits from the CVRPEnv class in which customers are considered.
    Additionally considers time windows within which a service has to be started.

    Observations:
        - location of the depot.
        - locations and demand of each customer.
        - current location of the vehicle.
        - the remaining customer of the vehicle.
        - the current time.
        - service durations of each location.
        - time windows of each location.

    Constraints:
        - the tour starts and ends at the depot.
        - each customer must be visited exactly once.
        - the vehicle cannot visit customers exceed the remaining customer.
        - the vehicle can return to the depot to refill the customer.
        - the vehicle must start the service within the time window of each location.

    Finish Condition:
        - the vehicle has visited all customers and returned to the depot.

    Reward:
        - (minus) the negative length of the path.

    Args:
        generator: CVRPTWGenerator instance as the data generator
        generator_params: parameters for the generator
    """

    name = "cvrptw"

    def __init__(
        self,
        generator: CVRPTWGenerator = None,
        generator_params: dict = {},
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = CVRPTWGenerator(**generator_params)
        self.generator = generator
        self._make_spec(self.generator)

    def _make_spec(self, generator: CVRPTWGenerator):
        if isinstance(generator, CVRPGenerator):
            super()._make_spec(generator)
        else:
            current_time = Unbounded(shape=(1), dtype=torch.float32, device=self.device)
            current_loc = Unbounded(shape=(2), dtype=torch.float32, device=self.device)
            durations = Bounded(
                low=generator.min_time,
                high=generator.max_time,
                shape=(generator.num_loc, 1),
                dtype=torch.int64,
                device=self.device,
            )
            time_windows = Bounded(
                low=generator.min_time,
                high=generator.max_time,
                shape=(
                    generator.num_loc,
                    2,
                ),  # Each location has a 2D time window (start, end)
                dtype=torch.int64,
                device=self.device,
            )
            # Extend observation specs
            self.observation_spec = Composite(
                **self.observation_spec,
                current_time=current_time,
                current_loc=current_loc,
                durations=durations,
                time_windows=time_windows,
            )

    @staticmethod
    def get_action_mask(td: TensorDict) -> torch.Tensor:
        """In addition to the constraints considered in the CVRPEnv, the time windows are considered.
        The vehicle can only visit a location if it can reach it in time, i.e. before its time window ends.
        """
        not_masked = CVRPEnv.get_action_mask(td)
        current_loc = gather_by_index(td["locs"], td["current_node"])
        dist = get_distance(current_loc[..., None, :], td["locs"])
        td.update({"current_loc": current_loc, "distances": dist})
        can_reach_in_time = (
            td["current_time"] + dist <= td["time_windows"][..., 1]
        )  # I only need to start the service before the time window ends, not finish it.
        return not_masked & can_reach_in_time

    def _step(self, td: TensorDict) -> TensorDict:
        """In addition to the calculations in the CVRPEnv, the current time is
        updated to keep track of which nodes are still reachable in time.
        The current_node is updeted in the parent class' _step() function.
        """
        batch_size = td["locs"].shape[0]
        # update current_time
        distance = gather_by_index(td["distances"], td["action"]).reshape([batch_size, 1])
        duration = gather_by_index(td["durations"], td["action"]).reshape([batch_size, 1])
        start_times = gather_by_index(td["time_windows"], td["action"])[..., 0].reshape(
            [batch_size, 1]
        )
        td["current_time"] = (td["action"][:, None] != 0) * (
            torch.max(td["current_time"] + distance, start_times) + duration
        )
        # current_node is updated to the selected action
        td = super()._step(td)
        return td

    def _reset(
        self, td: Optional[TensorDict] = None, batch_size: Optional[list] = None
    ) -> TensorDict:
        device = td.device
        td_reset = TensorDict(
            {
                "locs": torch.cat((td["depot"][..., None, :], td["locs"]), -2),
                "demand": td["demand"],
                "current_node": torch.zeros(
                    *batch_size, 1, dtype=torch.long, device=device
                ),
                "current_time": torch.zeros(
                    *batch_size, 1, dtype=torch.float32, device=device
                ),
                "used_capacity": torch.zeros((*batch_size, 1), device=device),
                "vehicle_capacity": torch.full(
                    (*batch_size, 1), self.generator.vehicle_capacity, device=device
                ),
                "visited": torch.zeros(
                    (*batch_size, td["locs"].shape[-2] + 1),
                    dtype=torch.uint8,
                    device=device,
                ),
                "durations": td["durations"],
                "time_windows": td["time_windows"],
            },
            batch_size=batch_size,
        )
        td_reset.set("action_mask", self.get_action_mask(td_reset))
        return td_reset

    def _get_reward(self, td: TensorDict, actions: torch.Tensor) -> torch.Tensor:
        """The reward is the negative tour length. Time windows
        are not considered for the calculation of the reward."""
        return super()._get_reward(td, actions)

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor) -> None:
        CVRPEnv.check_solution_validity(td, actions)
        batch_size = td["locs"].shape[0]
        # distances to depot
        distances = get_distance(
            td["locs"][..., 0, :], td["locs"].transpose(0, 1)
        ).transpose(0, 1)
        # basic checks on time windows
        assert torch.all(distances >= 0.0), "Distances must be non-negative."
        assert torch.all(td["time_windows"] >= 0.0), "Time windows must be non-negative."
        assert torch.all(
            td["time_windows"][..., :, 0] + distances + td["durations"]
            <= td["time_windows"][..., 0, 1][0]  # max_time is the same for all batches
        ), "vehicle cannot perform service and get back to depot in time."
        assert torch.all(
            td["durations"] >= 0.0
        ), "Service durations must be non-negative."
        assert torch.all(
            td["time_windows"][..., 0] < td["time_windows"][..., 1]
        ), "there are unfeasible time windows"
        # check vehicles can meet deadlines
        curr_time = torch.zeros(batch_size, 1, dtype=torch.float32, device=td.device)
        curr_node = torch.zeros_like(curr_time, dtype=torch.int64, device=td.device)
        for ii in range(actions.size(1)):
            next_node = actions[:, ii]
            dist = get_distance(
                gather_by_index(td["locs"], curr_node).reshape([batch_size, 2]),
                gather_by_index(td["locs"], next_node).reshape([batch_size, 2]),
            ).reshape([batch_size, 1])
            curr_time = torch.max(
                (curr_time + dist).int(),
                gather_by_index(td["time_windows"], next_node)[..., 0].reshape(
                    [batch_size, 1]
                ),
            )
            assert torch.all(
                curr_time
                <= gather_by_index(td["time_windows"], next_node)[..., 1].reshape(
                    [batch_size, 1]
                )
            ), "vehicle cannot start service before deadline"
            curr_time = curr_time + gather_by_index(td["durations"], next_node).reshape(
                [batch_size, 1]
            )
            curr_node = next_node
            curr_time[curr_node == 0] = 0.0  # reset time for depot

    @staticmethod
    def render(td: TensorDict, actions: torch.Tensor = None, ax=None):
        return render(td, actions, ax)

    @staticmethod
    def load_data(
        name: str,
        solomon=False,
        path_instances: str = None,
        type: str = None,
        compute_edge_weights: bool = False,
    ):
        if solomon:
            assert type in [
                "instance",
                "solution",
            ], "type must be either 'instance' or 'solution'"
            if type == "instance":
                instance = load_solomon_instance(
                    name=name, path=path_instances, edge_weights=compute_edge_weights
                )
            elif type == "solution":
                instance = load_solomon_solution(name=name, path=path_instances)
            return instance
        return load_npz_to_tensordict(filename=name)

    def extract_from_solomon(self, instance: dict, batch_size: int = 1):
        # extract parameters for the environment from the Solomon instance
        self.min_demand = instance["demand"][1:].min()
        self.max_demand = instance["demand"][1:].max()
        self.vehicle_capacity = instance["capacity"]
        self.min_loc = instance["node_coord"][1:].min()
        self.max_loc = instance["node_coord"][1:].max()
        self.min_time = instance["time_window"][:, 0].min()
        self.max_time = instance["time_window"][:, 1].max()
        # assert the time window of the depot starts at 0 and ends at max_time
        assert self.min_time == 0, "Time window of depot must start at 0."
        assert (
            self.max_time == instance["time_window"][0, 1]
        ), "Depot must have latest end time."
        # convert to format used in CVRPTWEnv
        td = TensorDict(
            {
                "depot": torch.tensor(
                    instance["node_coord"][0],
                    dtype=torch.float32,
                    device=self.device,
                ).repeat(batch_size, 1),
                "locs": torch.tensor(
                    instance["node_coord"][1:],
                    dtype=torch.float32,
                    device=self.device,
                ).repeat(batch_size, 1, 1),
                "demand": torch.tensor(
                    instance["demand"][1:],
                    dtype=torch.float32,
                    device=self.device,
                ).repeat(batch_size, 1),
                "durations": torch.tensor(
                    instance["service_time"],
                    dtype=torch.int64,
                    device=self.device,
                ).repeat(batch_size, 1),
                "time_windows": torch.tensor(
                    instance["time_window"],
                    dtype=torch.int64,
                    device=self.device,
                ).repeat(batch_size, 1, 1),
            },
            batch_size=1,  # we assume batch_size will always be 1 for loaded instances
        )
        return self.reset(td, batch_size=batch_size)



import torch

try:
    # from fla.ops.linear_attn.chunk_fuse import fused_chunk_linear_attn
    from fla.ops.linear_attn.chunk import chunk_linear_attn as fused_chunk_linear_attn
except ImportError:
    fused_chunk_linear_attn = None

try:
    from flash_attn import flash_attn_func
except ImportError:
    flash_attn_func = None


def fused_chunk_linear_attn_wrapper(
    q: torch.Tensor,
    k: torch.Tensor,
    v: torch.Tensor,
    scale: float = -1,
    initial_state: torch.Tensor = None,
    output_final_state: bool = False,
    normalize: bool = True,
    **kwargs,
):
    assert (
        fused_chunk_linear_attn is not None
    ), "fused_chunk_linear_attn not found. Install Flash Linear Attention using instructions from https://github.com/sustcsonglin/flash-linear-attention"
    assert (
        kwargs.get("attn_mask", None) is None
    ), "attn_mask is not supported in Flash  Linear Attention"
    return fused_chunk_linear_attn(
        q, k, v, scale, initial_state, output_final_state, normalize
    )[0]


def scaled_dot_product_attention_flash_attn(
    q, k, v, attn_mask=None, dropout_p=0.0, is_causal=False
):
    """
    Flash Attention 2 wrapper (https://github.com/Dao-AILab/flash-attention) around `flash_attn_func` to obtain the same behavior as
    `torch.nn.functional.scaled_dot_product_attention`.
    We need to permute the query, key, and value tensors before calling the scaled dot product attention function
    Reference: https://github.com/Dao-AILab/flash-attention/issues/383

    Note:
        Flash Attention does not support masking except for causal masking.

    Args:
        q (torch.Tensor): Query tensor of shape `(batch_size, num_heads, seq_len_q, head_dim)`
        k (torch.Tensor): Key tensor of shape `(batch_size, num_heads, seq_len_k, head_dim)`
        v (torch.Tensor): Value tensor of shape `(batch_size, num_heads, seq_len_v, head_dim)`
        attn_mask (torch.Tensor): Attention mask of shape `(batch_size, seq_len_q, seq_len_k)`
        dropout_p (float): Dropout probability
        is_causal (bool): Whether to apply causal mask to attention scores
    """
    assert attn_mask is None, "`attn_mask` is not supported in Flash Attention"
    assert flash_attn_func is not None, (
        "Flash Attention not found. Install Flash Attention using instructions from "
        "https://github.com/Dao-AILab/flash-attention . "
        "Alternatively, use `torch.nn.functional.scaled_dot_product_attention` available from PyTorch 2.0.0"
    )
    q, k, v = q.transpose(-2, -3), k.transpose(-2, -3), v.transpose(-2, -3)
    out = flash_attn_func(q, k, v, dropout_p=dropout_p, causal=is_causal)
    return out.transpose(-2, -3)



from typing import Callable

import torch

from tensordict.tensordict import TensorDict
from torch.distributions import Uniform

from rl4co.envs.common.utils import Generator, get_sampler
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class SVRPGenerator(Generator):
    """Data generator for the Skill Vehicle Routing Problem (SVRP).
    Args:
        num_loc: number of locations (customers) in the TSP
        min_loc: minimum value for the location coordinates
        max_loc: maximum value for the location coordinates
        loc_distribution: distribution for the location coordinates
        depot_distribution: distribution for the depot location. If None, sample the depot from the locations        min_skill: minimum value for the technic skill
        max_skill: maximum value for the technic skill
        skill_distribution: distribution for the technic skill
        tech_costs: list of the technic costs

    Returns:
        A TensorDict with the following keys:
            locs [batch_size, num_loc, 2]: locations of each customer
            depot [batch_size, 2]: location of the depot
            techs [batch_size, num_loc]: technic requirements of each customer
            skills [batch_size, num_loc]: skills of the vehicles
    """

    def __init__(
        self,
        num_loc: int = 20,
        min_loc: float = 0.0,
        max_loc: float = 1.0,
        loc_distribution: int | float | str | type | Callable = Uniform,
        depot_distribution: int | float | str | type | Callable = None,
        min_skill: float = 1.0,
        max_skill: float = 10.0,
        tech_costs: list = [1, 2, 3],
        **kwargs,
    ):
        self.num_loc = num_loc
        self.min_loc = min_loc
        self.max_loc = max_loc
        self.min_skill = min_skill
        self.max_skill = max_skill
        self.num_tech = len(tech_costs)
        self.tech_costs = torch.tensor(tech_costs)

        # Location distribution
        if kwargs.get("loc_sampler", None) is not None:
            self.loc_sampler = kwargs["loc_sampler"]
        else:
            self.loc_sampler = get_sampler(
                "loc", loc_distribution, min_loc, max_loc, **kwargs
            )

        # Depot distribution
        if kwargs.get("depot_sampler", None) is not None:
            self.depot_sampler = kwargs["depot_sampler"]
        else:
            self.depot_sampler = (
                get_sampler("depot", depot_distribution, min_loc, max_loc, **kwargs)
                if depot_distribution is not None
                else None
            )

    def _generate(self, batch_size) -> TensorDict:
        """Generate data for the basic Skill-VRP. The data consists of the locations of the customers,
        the skill-levels of the technicians and the required skill-levels of the customers.
        The data is generated randomly within the given bounds."""
        # Sample locations: depot and customers
        if self.depot_sampler is not None:
            depot = self.depot_sampler.sample((*batch_size, 2))
            locs = self.loc_sampler.sample((*batch_size, self.num_loc, 2))
        else:
            # if depot_sampler is None, sample the depot from the locations
            locs = self.loc_sampler.sample((*batch_size, self.num_loc + 1, 2))
            depot = locs[..., 0, :]
            locs = locs[..., 1:, :]

        locs_with_depot = torch.cat((depot[:, None, :], locs), dim=1)

        # Initialize technicians and sort ascendingly
        techs, _ = torch.sort(
            torch.FloatTensor(*batch_size, self.num_tech, 1).uniform_(
                self.min_skill, self.max_skill
            ),
            dim=-2,
        )

        # Initialize the skills
        skills = torch.FloatTensor(*batch_size, self.num_loc, 1).uniform_(0, 1)
        # scale skills
        skills = torch.max(techs, dim=1, keepdim=True).values * skills
        td = TensorDict(
            {
                "locs": locs_with_depot[..., 1:, :],
                "depot": locs_with_depot[..., 0, :],
                "techs": techs,
                "skills": skills,
            },
            batch_size=batch_size,
        )
        return td



from typing import Callable

import torch

from tensordict.tensordict import TensorDict
from torch.distributions import Uniform

from rl4co.envs.common.utils import Generator, get_sampler
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)

# From Kool et al. 2019
MAX_LENGTHS = {20: 2.0, 50: 3.0, 100: 4.0}


class OPGenerator(Generator):
    """Data generator for the Orienteering Problem (OP).

    Args:
        num_loc: number of locations (customers) in the OP, without the depot. (e.g. 10 means 10 locs + 1 depot)
        min_loc: minimum value for the location coordinates
        max_loc: maximum value for the location coordinates
        loc_distribution: distribution for the location coordinates
        depot_distribution: distribution for the depot location. If None, sample the depot from the locations
        min_prize: minimum value for the prize of each customer
        max_prize: maximum value for the prize of each customer
        prize_distribution: distribution for the prize of each customer
        max_length: maximum length of the path

    Returns:
        A TensorDict with the following keys:
            locs [batch_size, num_loc, 2]: locations of each customer
            depot [batch_size, 2]: location of the depot
            prize [batch_size, num_loc]: prize of each customer
            max_length [batch_size, 1]: maximum length of the path for each customer
    """

    def __init__(
        self,
        num_loc: int = 20,
        min_loc: float = 0.0,
        max_loc: float = 1.0,
        loc_distribution: int | float | str | type | Callable = Uniform,
        depot_distribution: int | float | str | type | Callable = None,
        min_prize: float = 1.0,
        max_prize: float = 1.0,
        prize_distribution: int | float | type | Callable = Uniform,
        prize_type: str = "dist",
        max_length: float | torch.Tensor = None,
        **kwargs,
    ):
        self.num_loc = num_loc
        self.min_loc = min_loc
        self.max_loc = max_loc
        self.min_prize = min_prize
        self.max_prize = max_prize
        self.prize_type = prize_type
        self.max_length = max_length

        # Location distribution
        if kwargs.get("loc_sampler", None) is not None:
            self.loc_sampler = kwargs["loc_sampler"]
        else:
            self.loc_sampler = get_sampler(
                "loc", loc_distribution, min_loc, max_loc, **kwargs
            )

        # Depot distribution
        if kwargs.get("depot_sampler", None) is not None:
            self.depot_sampler = kwargs["depot_sampler"]
        else:
            self.depot_sampler = (
                get_sampler("depot", depot_distribution, min_loc, max_loc, **kwargs)
                if depot_distribution is not None
                else None
            )

        # Prize distribution
        if kwargs.get("prize_sampler", None) is not None:
            self.prize_sampler = kwargs["prize_sampler"]
        elif (
            prize_distribution == "dist"
        ):  # If prize_distribution is 'dist', then the prize is the distance from the depot
            self.prize_sampler = None
        else:
            self.prize_sampler = get_sampler(
                "prize", prize_distribution, min_prize, max_prize, **kwargs
            )

        # Max length
        if max_length is not None:
            self.max_length = max_length
        else:
            self.max_length = MAX_LENGTHS.get(num_loc, None)
        if self.max_length is None:
            closest_num_loc = min(MAX_LENGTHS.keys(), key=lambda x: abs(x - num_loc))
            self.max_length = MAX_LENGTHS[closest_num_loc]
            log.warning(
                f"The max length for {num_loc} locations is not defined. Using the closest max length: {self.max_length}\
                    with {closest_num_loc} locations."
            )

    def _generate(self, batch_size) -> TensorDict:
        # Sample locations: depot and customers
        if self.depot_sampler is not None:
            depot = self.depot_sampler.sample((*batch_size, 2))
            locs = self.loc_sampler.sample((*batch_size, self.num_loc, 2))
        else:
            # if depot_sampler is None, sample the depot from the locations
            locs = self.loc_sampler.sample((*batch_size, self.num_loc + 1, 2))
            depot = locs[..., 0, :]
            locs = locs[..., 1:, :]

        locs_with_depot = torch.cat((depot.unsqueeze(1), locs), dim=1)

        # Methods taken from Fischetti et al. (1998) and Kool et al. (2019)
        if self.prize_type == "const":
            prize = torch.ones(*batch_size, self.num_loc, device=self.device)
        elif self.prize_type == "unif":
            prize = (
                1
                + torch.randint(
                    0, 100, (*batch_size, self.num_loc), device=self.device
                ).float()
            ) / 100
        elif self.prize_type == "dist":  # based on the distance to the depot
            prize = (locs_with_depot[..., 0:1, :] - locs_with_depot[..., 1:, :]).norm(
                p=2, dim=-1
            )
            prize = (
                1 + (prize / prize.max(dim=-1, keepdim=True)[0] * 99).int()
            ).float() / 100
        else:
            raise ValueError(f"Invalid prize_type: {self.prize_type}")

        # Support for heterogeneous max length if provided
        if not isinstance(self.max_length, torch.Tensor):
            max_length = torch.full((*batch_size,), self.max_length)
        else:
            max_length = self.max_length

        return TensorDict(
            {
                "locs": locs_with_depot[..., 1:, :],
                "depot": locs_with_depot[..., 0, :],
                "prize": prize,
                "max_length": max_length,
            },
            batch_size=batch_size,
        )



import torch.nn as nn

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.zoo.matnet.policy import MatNetPolicy, MultiStageFFSPPolicy
from rl4co.models.zoo.pomo import POMO
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def select_matnet_policy(env, **policy_params):
    if env.name == "ffsp":
        if env.flatten_stages:
            return MatNetPolicy(env_name=env.name, **policy_params)
        else:
            return MultiStageFFSPPolicy(stage_cnt=env.num_stage, **policy_params)
    else:
        return MatNetPolicy(env_name=env.name, **policy_params)


class MatNet(POMO):
    def __init__(
        self,
        env: RL4COEnvBase,
        policy: nn.Module | MatNetPolicy = None,
        num_starts: int = None,
        policy_params: dict = {},
        **kwargs,
    ):
        if policy is None:
            policy = select_matnet_policy(env=env, **policy_params)

        # Check if using augmentation and the validation of augmentation function
        if kwargs.get("num_augment", 0) != 0:
            log.warning("MatNet is using augmentation.")
            if (
                kwargs.get("augment_fn") in ["symmetric", "dihedral8"]
                or kwargs.get("augment_fn") is None
            ):
                log.error(
                    "MatNet does not use symmetric or dihedral augmentation. Seeting no augmentation function."
                )
                kwargs["num_augment"] = 0
        else:
            kwargs["num_augment"] = 0

        super(MatNet, self).__init__(
            env=env,
            policy=policy,
            num_starts=num_starts,
            **kwargs,
        )



import lightning.pytorch as pl
import torch
import math
import copy
from torch.optim import Adam

from lightning import Callback
from rl4co import utils
import random
log = utils.get_pylogger(__name__)


class ReptileCallback(Callback):

    """ Meta training framework for addressing the generalization issue (implement the Reptile algorithm only)
    Based on Manchanda et al. 2022 (https://arxiv.org/abs/2206.00787) and Zhou et al. 2023 (https://arxiv.org/abs/2305.19587)

    Args:
        - num_tasks: the number of tasks in a mini-batch, i.e. `B` in the original paper
        - alpha: initial weight of the task model for the outer-loop optimization of reptile
        - alpha_decay: weight decay of the task model for the outer-loop optimization of reptile
        - min_size: minimum problem size of the task (only supported in cross-size generalization)
        - max_size: maximum problem size of the task (only supported in cross-size generalization)
        - sch_bar: for the task scheduler of size setting, where lr_decay_epoch = sch_bar * epochs, i.e. after this epoch, learning rate will decay with a weight 0.1
        - data_type: type of the tasks, chosen from ["size", "distribution", "size_distribution"]
        - print_log: whether to print the specific task sampled in each inner-loop optimization
    """
    def __init__(self,
                 num_tasks: int,
                 alpha: float,
                 alpha_decay: float,
                 min_size: int,
                 max_size: int,
                 sch_bar: float = 0.9,
                 data_type: str = "size",
                 print_log: bool =True):

        super().__init__()

        self.num_tasks = num_tasks
        self.alpha = alpha
        self.alpha_decay = alpha_decay
        self.sch_bar = sch_bar
        self.print_log = print_log
        self.data_type = data_type
        self.task_set = self._generate_task_set(data_type, min_size, max_size)

    def on_fit_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:

        # Sample a batch of tasks
        self._sample_task()

        # Pre-set the distribution
        if self.data_type == "size_distribution":
            pl_module.env.generator.loc_distribution = "gaussian_mixture"
            self.selected_tasks[0] = (pl_module.env.generator.num_loc, 0, 0)
        elif self.data_type == "size":
            pl_module.env.generator.loc_distribution = "uniform"
            self.selected_tasks[0] = (pl_module.env.generator.num_loc, )
        elif self.data_type == "distribution":
            pl_module.env.generator.loc_distribution = "gaussian_mixture"
            self.selected_tasks[0] = (0, 0)
        self.task_params = self.selected_tasks[0]

    def on_train_epoch_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:

        # Alpha scheduler (decay for the update of meta model)
        self._alpha_scheduler()

        # Reinitialize the task model with the parameters of the meta model
        if trainer.current_epoch %  self.num_tasks == 0: # Save the meta model
            self.meta_model_state_dict = copy.deepcopy(pl_module.state_dict())
            self.task_models = []
            # Print sampled tasks
            if self.print_log:
                print('\n>> Meta epoch: {} (Exact epoch: {}), Training task: {}'.format(trainer.current_epoch//self.num_tasks, trainer.current_epoch, self.selected_tasks))
        else:
            pl_module.load_state_dict(self.meta_model_state_dict)

        # Reinitialize the optimizer every epoch
        lr_decay = 0.1 if trainer.current_epoch+1 == int(self.sch_bar * trainer.max_epochs) else 1
        old_lr  = trainer.optimizers[0].param_groups[0]['lr']
        new_optimizer = Adam(pl_module.parameters(), lr=old_lr * lr_decay)
        trainer.optimizers = [new_optimizer]

        # Print
        if self.print_log:
            if hasattr(pl_module.env.generator, 'capacity'):
                print('>> Training task: {}, capacity: {}'.format(self.task_params, pl_module.env.generator.capacity))
            else:
                print('>> Training task: {}'.format(self.task_params))

    def on_train_epoch_end(self,  trainer: pl.Trainer, pl_module: pl.LightningModule):

        # Save the task model
        self.task_models.append(copy.deepcopy(pl_module.state_dict()))
        if (trainer.current_epoch+1) % self.num_tasks == 0:
            # Outer-loop optimization (update the meta model with the parameters of the task model)
            with torch.no_grad():
                state_dict = {params_key: (self.meta_model_state_dict[params_key] +
                                           self.alpha * torch.mean(torch.stack([fast_weight[params_key] - self.meta_model_state_dict[params_key]
                                                                                for fast_weight in self.task_models], dim=0).float(), dim=0))
                              for params_key in self.meta_model_state_dict}
                pl_module.load_state_dict(state_dict)

        # Get ready for the next meta-training iteration
        if (trainer.current_epoch + 1) % self.num_tasks == 0:
            # Sample a batch of tasks
            self._sample_task()

        # Load new training task (Update the environment) for the next meta-training iteration
        self._load_task(pl_module, task_idx = (trainer.current_epoch+1) % self.num_tasks)

    def _sample_task(self):

        # Sample a batch of tasks
        self.selected_tasks = []
        for b in range(self.num_tasks):
            task_params = random.sample(self.task_set, 1)[0]
            self.selected_tasks.append(task_params)

    def _load_task(self, pl_module: pl.LightningModule, task_idx=0):

        # Load new training task (Update the environment)
        self.task_params = self.selected_tasks[task_idx]

        if self.data_type == "size_distribution":
            assert len(self.task_params) == 3
            pl_module.env.generator.num_loc = self.task_params[0]
            pl_module.env.generator.num_modes = self.task_params[1]
            pl_module.env.generator.cdist = self.task_params[2]
        elif self.data_type == "distribution": # fixed size
            assert len(self.task_params) == 2
            pl_module.env.generator.num_modes = self.task_params[0]
            pl_module.env.generator.cdist = self.task_params[1]
        elif self.data_type == "size": # fixed distribution
            assert len(self.task_params) == 1
            pl_module.env.generator.num_loc = self.task_params[0]

        if hasattr(pl_module.env.generator, 'capacity') and self.data_type in ["size_distribution", "size"]:
            task_capacity = math.ceil(30 + self.task_params[0] / 5) if self.task_params[0] >= 20 else 20
            pl_module.env.generator.capacity = task_capacity

    def _alpha_scheduler(self):
        self.alpha = max(self.alpha * self.alpha_decay, 0.0001)

    def _generate_task_set(self, data_type, min_size, max_size):
        """
        Following the setting in Zhou et al. 2023 (https://arxiv.org/abs/2305.19587)
        Current setting:
            size: (n,) \in [20, 150]
            distribution: (m, c) \in {(0, 0) + [1-9] * [1, 10, 20, 30, 40, 50]}
            size_distribution: (n, m, c) \in [50, 200, 5] * {(0, 0) + (1, 1) + [3, 5, 7] * [10, 30, 50]}
        """

        if data_type == "distribution":  # focus on TSP100 with gaussian mixture distributions
            task_set = [(0, 0)] + [(m, c) for m in range(1, 10) for c in [1, 10, 20, 30, 40, 50]]
        elif data_type == "size":  # focus on uniform distribution with different sizes
            task_set = [(n,) for n in range(min_size, max_size + 1)]
        elif data_type == "size_distribution":
            dist_set = [(0, 0), (1, 1)] + [(m, c) for m in [3, 5, 7] for c in [10, 30, 50]]
            task_set = [(n, m, c) for n in range(50, 201, 5) for (m, c) in dist_set]
        else:
            raise NotImplementedError

        print(">> Generating training task set: {} tasks with type {}".format(len(task_set), data_type))
        print(">> Training task set: {}".format(task_set))

        return task_set




from typing import Optional

import torch

from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, Unbounded

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.envs.common.utils import batch_to_scalar
from rl4co.utils.pylogger import get_pylogger

from .generator import ATSPGenerator
from .render import render

log = get_pylogger(__name__)


class ATSPEnv(RL4COEnvBase):
    """Asymmetric Traveling Salesman Problem (ATSP) environment
    At each step, the agent chooses a customer to visit. The reward is 0 unless the agent visits all the customers.
    In that case, the reward is (-)length of the path: maximizing the reward is equivalent to minimizing the path length.
    Unlike the TSP, the distance matrix is asymmetric, i.e., the distance from A to B is not necessarily the same as the distance from B to A.

    Observations:
        - distance matrix between customers
        - the current customer
        - the first customer (for calculating the reward)
        - the remaining unvisited customers

    Constraints:
        - the tour starts and ends at the same customer.
        - each customer must be visited exactly once.

    Finish Condition:
        - the agent has visited all customers.

    Reward:
        - (minus) the negative length of the path.

    Args:
        generator: ATSPGenerator instance as the data generator
        generator_params: parameters for the generator
    """

    name = "atsp"

    def __init__(
        self,
        generator: ATSPGenerator = None,
        generator_params: dict = {},
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = ATSPGenerator(**generator_params)
        self.generator = generator
        self._make_spec(self.generator)

    @staticmethod
    def _step(td: TensorDict) -> TensorDict:
        current_node = td["action"]
        first_node = current_node if batch_to_scalar(td["i"]) == 0 else td["first_node"]

        # Set not visited to 0 (i.e., we visited the node)
        available = td["action_mask"].scatter(
            -1, current_node.unsqueeze(-1).expand_as(td["action_mask"]), 0
        )

        # We are done there are no unvisited locations
        done = torch.count_nonzero(available, dim=-1) <= 0

        # The reward is calculated outside via get_reward for efficiency, so we set it to 0 here
        reward = torch.zeros_like(done)

        td.update(
            {
                "first_node": first_node,
                "current_node": current_node,
                "i": td["i"] + 1,
                "action_mask": available,
                "reward": reward,
                "done": done,
            },
        )
        return td

    def _reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        # Initialize distance matrix
        cost_matrix = td["cost_matrix"]
        device = td.device

        # Other variables
        current_node = torch.zeros((*batch_size, 1), dtype=torch.int64, device=device)
        available = torch.ones(
            (*batch_size, self.generator.num_loc), dtype=torch.bool, device=device
        )  # 1 means not visited, i.e. action is allowed
        i = torch.zeros((*batch_size, 1), dtype=torch.int64, device=device)

        return TensorDict(
            {
                "cost_matrix": cost_matrix,
                "first_node": current_node,
                "current_node": current_node,
                "i": i,
                "action_mask": available,
            },
            batch_size=batch_size,
        )

    def _make_spec(self, generator: ATSPGenerator):
        self.observation_spec = Composite(
            cost_matrix=Bounded(
                low=generator.min_dist,
                high=generator.max_dist,
                shape=(generator.num_loc, generator.num_loc),
                dtype=torch.float32,
            ),
            first_node=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            current_node=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            i=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            action_mask=Unbounded(
                shape=(generator.num_loc),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            shape=(1,),
            dtype=torch.int64,
            low=0,
            high=generator.num_loc,
        )
        self.reward_spec = Unbounded(shape=(1,))
        self.done_spec = Unbounded(shape=(1,), dtype=torch.bool)

    def _get_reward(self, td: TensorDict, actions: torch.Tensor) -> torch.Tensor:
        distance_matrix = td["cost_matrix"]

        # Get indexes of tour edges
        nodes_src = actions
        nodes_tgt = torch.roll(actions, -1, dims=1)
        batch_idx = torch.arange(
            distance_matrix.shape[0], device=distance_matrix.device
        ).unsqueeze(1)
        # return negative tour length
        return -distance_matrix[batch_idx, nodes_src, nodes_tgt].sum(-1)

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor):
        assert (
            torch.arange(actions.size(1), out=actions.data.new())
            .view(1, -1)
            .expand_as(actions)
            == actions.data.sort(1)[0]
        ).all(), "Invalid tour"

    @staticmethod
    def render(td, actions=None, ax=None):
        return render(td, actions, ax)



import numpy as np
import torch

from tensordict import TensorDict
from torch import Tensor


def process_instance(td: TensorDict) -> TensorDict:
    """
    We simply transform the data to the format the current PyVRP API expects
    """
    td_ = td.clone().cpu()
    td_.set("durations", td["service_time"])
    cost_mat = torch.cdist(td_["locs"], td_["locs"])
    num_loc = cost_mat.shape[-1]
    # note: if we don't do this, PyVRP may complain diagonal is not 0.
    # i guess it is because of some conversion from floating point to integer
    cost_mat[:, torch.arange(num_loc), torch.arange(num_loc)] = 0
    td_.set("cost_matrix", cost_mat)
    backhaul_class = td.get("backhaul_class", torch.ones(td_.batch_size[0], 1))
    td_.set("backhaul_class", backhaul_class)
    return td_


def scale(data: Tensor, scaling_factor: int):
    """
    Scales ands rounds data to integers so PyVRP can handle it.
    """
    array = (data * scaling_factor).numpy().round()
    array = np.where(array == np.inf, np.iinfo(np.int32).max, array)
    array = array.astype(int)

    if array.size == 1:
        return array.item()

    return array



import time

import numpy as np
import torch

from torch.utils.data import DataLoader
from tqdm.auto import tqdm

from rl4co.data.transforms import StateAugmentation
from rl4co.utils.ops import batchify, gather_by_index, sample_n_random_actions, unbatchify


def check_unused_kwargs(class_, kwargs):
    if len(kwargs) > 0 and not (len(kwargs) == 1 and "progress" in kwargs):
        print(f"Warning: {class_.__class__.__name__} does not use kwargs {kwargs}")


class EvalBase:
    """Base class for evaluation

    Args:
        env: Environment
        progress: Whether to show progress bar
        **kwargs: Additional arguments (to be implemented in subclasses)
    """

    name = "base"

    def __init__(self, env, progress=True, **kwargs):
        check_unused_kwargs(self, kwargs)
        self.env = env
        self.progress = progress

    def __call__(self, policy, dataloader, **kwargs):
        """Evaluate the policy on the given dataloader with **kwargs parameter
        self._inner is implemented in subclasses and returns actions and rewards
        """
        start = time.time()

        with torch.inference_mode():
            rewards_list = []
            actions_list = []

            for batch in tqdm(
                dataloader, disable=not self.progress, desc=f"Running {self.name}"
            ):
                td = batch.to(next(policy.parameters()).device)
                td = self.env.reset(td)
                actions, rewards = self._inner(policy, td, **kwargs)
                rewards_list.append(rewards)
                actions_list.append(actions)

            rewards = torch.cat(rewards_list)

            # Padding: pad actions to the same length with zeros
            max_length = max(action.size(-1) for action in actions_list)
            actions = torch.cat(
                [
                    torch.nn.functional.pad(action, (0, max_length - action.size(-1)))
                    for action in actions_list
                ],
                0,
            )

        inference_time = time.time() - start

        tqdm.write(f"Mean reward for {self.name}: {rewards.mean():.4f}")
        tqdm.write(f"Time: {inference_time:.4f}s")

        # Empty cache
        if torch.cuda.is_available():
            torch.cuda.empty_cache()

        return {
            "actions": actions.cpu(),
            "rewards": rewards.cpu(),
            "inference_time": inference_time,
            "avg_reward": rewards.cpu().mean(),
        }

    def _inner(self, policy, td):
        """Inner function to be implemented in subclasses.
        This function returns actions and rewards for the given policy
        """
        raise NotImplementedError("Implement in subclass")


class GreedyEval(EvalBase):
    """Evaluates the policy using greedy decoding and single trajectory"""

    name = "greedy"

    def __init__(self, env, **kwargs):
        check_unused_kwargs(self, kwargs)
        super().__init__(env, kwargs.get("progress", True))

    def _inner(self, policy, td):
        out = policy(
            td.clone(),
            decode_type="greedy",
            num_starts=0,
        )
        rewards = self.env.get_reward(td, out["actions"])
        return out["actions"], rewards


class AugmentationEval(EvalBase):
    """Evaluates the policy via N state augmentations
    `force_dihedral_8` forces the use of 8 augmentations (rotations and flips) as in POMO
    https://en.wikipedia.org/wiki/Examples_of_groups#dihedral_group_of_order_8

    Args:
        num_augment (int): Number of state augmentations
        force_dihedral_8 (bool): Whether to force the use of 8 augmentations
    """

    name = "augmentation"

    def __init__(self, env, num_augment=8, force_dihedral_8=False, feats=None, **kwargs):
        check_unused_kwargs(self, kwargs)
        super().__init__(env, kwargs.get("progress", True))
        self.augmentation = StateAugmentation(
            num_augment=num_augment,
            augment_fn="dihedral8" if force_dihedral_8 else "symmetric",
            feats=feats,
        )

    def _inner(self, policy, td, num_augment=None):
        if num_augment is None:
            num_augment = self.augmentation.num_augment
        td_init = td.clone()
        td = self.augmentation(td)
        out = policy(td.clone(), decode_type="greedy", num_starts=0)

        # Move into batches and compute rewards
        rewards = self.env.get_reward(batchify(td_init, num_augment), out["actions"])
        rewards = unbatchify(rewards, num_augment)
        actions = unbatchify(out["actions"], num_augment)

        # Get best reward and corresponding action
        rewards, max_idxs = rewards.max(dim=1)
        actions = gather_by_index(actions, max_idxs, dim=1)
        return actions, rewards

    @property
    def num_augment(self):
        return self.augmentation.num_augment


class SamplingEval(EvalBase):
    """Evaluates the policy via N samples from the policy

    Args:
        samples (int): Number of samples to take
        softmax_temp (float): Temperature for softmax sampling. The higher the temperature, the more random the sampling
    """

    name = "sampling"

    def __init__(
        self,
        env,
        samples,
        softmax_temp=None,
        select_best=True,
        temperature=1.0,
        top_p=0.0,
        top_k=0,
        **kwargs,
    ):
        check_unused_kwargs(self, kwargs)
        super().__init__(env, kwargs.get("progress", True))

        self.samples = samples
        self.softmax_temp = softmax_temp
        self.temperature = temperature
        self.select_best = select_best
        self.top_p = top_p
        self.top_k = top_k

    def _inner(self, policy, td):
        out = policy(
            td.clone(),
            decode_type="sampling",
            num_starts=self.samples,
            temperature=self.temperature,
            top_p=self.top_p,
            top_k=self.top_k,
            multisample=True,
            softmax_temp=self.softmax_temp,
            select_best=self.select_best,
            select_start_nodes_fn=lambda td, _, n: sample_n_random_actions(td, n),
        )

        # Move into batches and compute rewards
        rewards = out["reward"]
        actions = out["actions"]

        return actions, rewards


class GreedyMultiStartEval(EvalBase):
    """Evaluates the policy via `num_starts` greedy multistarts samples from the policy

    Args:
        num_starts (int): Number of greedy multistarts to use
    """

    name = "multistart_greedy"

    def __init__(self, env, num_starts=None, **kwargs):
        check_unused_kwargs(self, kwargs)
        super().__init__(env, kwargs.get("progress", True))

        assert num_starts is not None, "Must specify num_starts"
        self.num_starts = num_starts

    def _inner(self, policy, td):
        td_init = td.clone()
        out = policy(
            td.clone(),
            decode_type="multistart_greedy",
            num_starts=self.num_starts,
        )

        # Move into batches and compute rewards
        td = batchify(td_init, self.num_starts)
        rewards = self.env.get_reward(td, out["actions"])
        rewards = unbatchify(rewards, self.num_starts)
        actions = unbatchify(out["actions"], self.num_starts)

        # Get the best trajectories
        rewards, max_idxs = rewards.max(dim=1)
        actions = gather_by_index(actions, max_idxs, dim=1)
        return actions, rewards


class GreedyMultiStartAugmentEval(EvalBase):
    """Evaluates the policy via `num_starts` samples from the policy
    and `num_augment` augmentations of each sample.`
    `force_dihedral_8` forces the use of 8 augmentations (rotations and flips) as in POMO
    https://en.wikipedia.org/wiki/Examples_of_groups#dihedral_group_of_order_8

    Args:
        num_starts: Number of greedy multistart samples
        num_augment: Number of augmentations per sample
        force_dihedral_8: If True, force the use of 8 augmentations (rotations and flips) as in POMO
    """

    name = "multistart_greedy_augment"

    def __init__(
        self,
        env,
        num_starts=None,
        num_augment=8,
        force_dihedral_8=False,
        feats=None,
        **kwargs,
    ):
        check_unused_kwargs(self, kwargs)
        super().__init__(env, kwargs.get("progress", True))

        assert num_starts is not None, "Must specify num_starts"
        self.num_starts = num_starts
        assert not (
            num_augment != 8 and force_dihedral_8
        ), "Cannot force dihedral 8 when num_augment != 8"
        self.augmentation = StateAugmentation(
            num_augment=num_augment,
            augment_fn="dihedral8" if force_dihedral_8 else "symmetric",
            feats=feats,
        )

    def _inner(self, policy, td, num_augment=None):
        if num_augment is None:
            num_augment = self.augmentation.num_augment

        td_init = td.clone()

        td = self.augmentation(td)
        out = policy(
            td.clone(),
            decode_type="multistart_greedy",
            num_starts=self.num_starts,
        )

        # Move into batches and compute rewards
        td = batchify(td_init, (num_augment, self.num_starts))
        rewards = self.env.get_reward(td, out["actions"])
        rewards = unbatchify(rewards, self.num_starts * num_augment)
        actions = unbatchify(out["actions"], self.num_starts * num_augment)

        # Get the best trajectories
        rewards, max_idxs = rewards.max(dim=1)
        actions = gather_by_index(actions, max_idxs, dim=1)
        return actions, rewards

    @property
    def num_augment(self):
        return self.augmentation.num_augment


def get_automatic_batch_size(eval_fn, start_batch_size=8192, max_batch_size=4096):
    """Automatically reduces the batch size based on the eval function

    Args:
        eval_fn: The eval function
        start_batch_size: The starting batch size. This should be the theoretical maximum batch size
        max_batch_size: The maximum batch size. This is the practical maximum batch size
    """
    batch_size = start_batch_size

    effective_ratio = 1

    if hasattr(eval_fn, "num_starts"):
        batch_size = batch_size // (eval_fn.num_starts // 10)
        effective_ratio *= eval_fn.num_starts // 10
    if hasattr(eval_fn, "num_augment"):
        batch_size = batch_size // eval_fn.num_augment
        effective_ratio *= eval_fn.num_augment
    if hasattr(eval_fn, "samples"):
        batch_size = batch_size // eval_fn.samples
        effective_ratio *= eval_fn.samples

    batch_size = min(batch_size, max_batch_size)
    # get closest integer power of 2
    batch_size = 2 ** int(np.log2(batch_size))

    print(f"Effective batch size: {batch_size} (ratio: {effective_ratio})")

    return batch_size


def evaluate_policy(
    env,
    policy,
    dataset,
    method="greedy",
    batch_size=None,
    max_batch_size=4096,
    start_batch_size=8192,
    auto_batch_size=True,
    samples=1280,
    softmax_temp=1.0,
    num_augment=8,
    force_dihedral_8=True,
    **kwargs,
):
    num_loc = getattr(env.generator, "num_loc", None)

    methods_mapping = {
        "greedy": {"func": GreedyEval, "kwargs": {}},
        "sampling": {
            "func": SamplingEval,
            "kwargs": {"samples": samples, "softmax_temp": softmax_temp},
        },
        "multistart_greedy": {
            "func": GreedyMultiStartEval,
            "kwargs": {"num_starts": num_loc},
        },
        "augment_dihedral_8": {
            "func": AugmentationEval,
            "kwargs": {"num_augment": num_augment, "force_dihedral_8": force_dihedral_8},
        },
        "augment": {"func": AugmentationEval, "kwargs": {"num_augment": num_augment}},
        "multistart_greedy_augment_dihedral_8": {
            "func": GreedyMultiStartAugmentEval,
            "kwargs": {
                "num_augment": num_augment,
                "force_dihedral_8": force_dihedral_8,
                "num_starts": num_loc,
            },
        },
        "multistart_greedy_augment": {
            "func": GreedyMultiStartAugmentEval,
            "kwargs": {"num_augment": num_augment, "num_starts": num_loc},
        },
    }

    assert method in methods_mapping, "Method {} not found".format(method)

    # Set up the evaluation function
    eval_settings = methods_mapping[method]
    func, kwargs_ = eval_settings["func"], eval_settings["kwargs"]
    # subsitute kwargs with the ones passed in
    kwargs_.update(kwargs)
    kwargs = kwargs_
    eval_fn = func(env, **kwargs)

    if auto_batch_size:
        assert (
            batch_size is None
        ), "Cannot specify batch_size when auto_batch_size is True"
        batch_size = get_automatic_batch_size(
            eval_fn, max_batch_size=max_batch_size, start_batch_size=start_batch_size
        )
        print("Using automatic batch size: {}".format(batch_size))

    # Set up the dataloader
    dataloader = DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=0,
        collate_fn=dataset.collate_fn,
    )

    # Run evaluation
    retvals = eval_fn(policy, dataloader)

    return retvals


if __name__ == "__main__":
    import argparse
    import importlib
    import os
    import pickle

    import torch

    from rl4co.envs import get_env

    parser = argparse.ArgumentParser()

    # Environment
    parser.add_argument("--problem", type=str, default="tsp", help="Problem to solve")
    parser.add_argument(
        "--generator-params",
        type=dict,
        default={"num_loc": 50},
        help="Generator parameters for the environment",
    )
    parser.add_argument(
        "--data-path",
        type=str,
        default="data/tsp/tsp50_test_seed1234.npz",
        help="Path of the test data npz file",
    )

    # Model
    parser.add_argument(
        "--model",
        type=str,
        default="AttentionModel",
        help="The class name of the valid model",
    )
    parser.add_argument(
        "--ckpt-path",
        type=str,
        default="checkpoints/am-tsp50.ckpt",
        help="The path of the checkpoint file",
    )
    parser.add_argument(
        "--device", type=str, default="cuda:1", help="Device to run the evaluation"
    )

    # Evaluation
    parser.add_argument(
        "--method",
        type=str,
        default="greedy",
        help="Evaluation method, support 'greedy', 'sampling',\
                        'multistart_greedy', 'augment_dihedral_8', 'augment', 'multistart_greedy_augment_dihedral_8',\
                        'multistart_greedy_augment'",
    )
    parser.add_argument(
        "--temperature", type=float, default=1.0, help="Temperature for sampling"
    )
    parser.add_argument(
        "--top-p",
        type=float,
        default=0.0,
        help="Top-p for sampling, from 0.0 to 1.0, 0.0 means not activated",
    )
    parser.add_argument("--top-k", type=int, default=0, help="Top-k for sampling")
    parser.add_argument(
        "--select-best",
        default=True,
        action=argparse.BooleanOptionalAction,
        help="During sampling, whether to select the best action, use --no-select_best to disable",
    )
    parser.add_argument(
        "--save-results",
        default=True,
        action=argparse.BooleanOptionalAction,
        help="Whether to save the evaluation results",
    )
    parser.add_argument(
        "--save-path",
        type=str,
        default="results",
        help="The root path to save the results",
    )
    parser.add_argument(
        "--num-instances",
        type=int,
        default=1000,
        help="Number of instances to test, maximum 10000",
    )

    parser.add_argument(
        "--samples", type=int, default=1280, help="Number of samples for sampling method"
    )
    parser.add_argument(
        "--softmax-temp",
        type=float,
        default=1.0,
        help="Temperature for softmax in the sampling method",
    )
    parser.add_argument(
        "--num-augment",
        type=int,
        default=8,
        help="Number of augmentations for augmentation method",
    )
    parser.add_argument(
        "--force-dihedral-8",
        default=True,
        action=argparse.BooleanOptionalAction,
        help="Force the use of 8 augmentations for augmentation method",
    )

    opts = parser.parse_args()

    # Log the evaluation setting information
    print(f"Problem: {opts.problem}-{opts.generator_params['num_loc']}")
    print(f"Model: {opts.model}")
    print(f"Loading test instances from: {opts.data_path}")
    print(f"Loading model checkpoint from: {opts.ckpt_path}")
    print(f"Using the device: {opts.device}")
    print(f"Evaluation method: {opts.method}")
    print(f"Number of instances to test: {opts.num_instances}")

    if opts.method == "sampling":
        print(f"[Sampling] Number of samples: {opts.samples}")
        print(f"[Sampling] Temperature: {opts.temperature}")
        print(f"[Sampling] Top-p: {opts.top_p}")
        print(f"[Sampling] Top-k: {opts.top_k}")
        print(f"[Sampling] Softmax temperature: {opts.softmax_temp}")
        print(f"[Sampling] Select best: {opts.select_best}")

    if opts.method == "augment" or opts.method == "augment_dihedral_8":
        print(f"[Augmentation] Number of augmentations: {opts.num_augment}")
        print(f"[Augmentation] Force dihedral 8: {opts.force_dihedral_8}")

    if opts.save_results:
        print(f"Saving the results to: {opts.save_path}")
    else:
        print("[Warning] The result will not be saved!")

    # Init the environment
    env = get_env(opts.problem, generator_params=opts.generator_params)

    # Load the test data
    dataset = env.dataset(filename=opts.data_path)

    # Restrict the instances of testing
    dataset.data_len = min(opts.num_instances, len(dataset))

    # Load the model from checkpoint
    model_root = importlib.import_module("rl4co.models.zoo")
    model_cls = getattr(model_root, opts.model)
    model = model_cls.load_from_checkpoint(opts.ckpt_path, load_baseline=False)
    model = model.to(opts.device)

    # Evaluate
    result = evaluate_policy(
        env=env,
        policy=model.policy,
        dataset=dataset,
        method=opts.method,
        temperature=opts.temperature,
        top_p=opts.top_p,
        top_k=opts.top_k,
        samples=opts.samples,
        softmax_temp=opts.softmax_temp,
        num_augment=opts.num_augment,
        select_best=True,
        force_dihedral_8=True,
    )

    # Save the results
    if opts.save_results:
        if not os.path.exists(opts.save_path):
            os.makedirs(opts.save_path)
        save_fname = f"{env.name}{env.generator.num_loc}-{opts.model}-{opts.method}-temp-{opts.temperature}-top_p-{opts.top_p}-top_k-{opts.top_k}.pkl"
        save_path = os.path.join(opts.save_path, save_fname)
        with open(save_path, "wb") as f:
            pickle.dump(result, f)



import matplotlib.pyplot as plt
import numpy as np
import torch

from rl4co.utils.ops import gather_by_index
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def render(td, actions=None, ax=None):

    if ax is None:
        # Create a plot of the nodes
        _, ax = plt.subplots(figsize=(3, 3))

    td = td.detach().cpu()

    if actions is None:
        actions = td.get("action", None)
    # if batch_size greater than 0 , we need to select the first batch element
    if td.batch_size != torch.Size([]):
        td = td[0]
        actions = actions[0]

    locs = td["locs"]

    # gather locs in order of action if available
    if actions is None:
        log.warning("No action in TensorDict, rendering unsorted locs")
    else:
        actions = actions.detach().cpu()
        locs = gather_by_index(locs, actions, dim=0)

    start_x, start_y = locs[0, 0], locs[0, 1]
    end_x, end_y = locs[-1, 0], locs[-1, 1]
    city_x, city_y = locs[1:-1, 0], locs[1:-1, 1]
    x, y = locs[:, 0], locs[:, 1]

    # Plot the start and end nodes
    ax.scatter(start_x, start_y, color="tab:green", marker="s")
    ax.scatter(end_x, end_y, color="tab:red", marker="x")

    # Plot the visited nodes
    ax.scatter(city_x, city_y, color="tab:blue")

    # Add arrows between visited nodes as a quiver plot
    dx, dy = np.diff(x), np.diff(y)
    ax.quiver(
        x[:-1],
        y[:-1],
        dx,
        dy,
        scale_units="xy",
        angles="xy",
        scale=1,
        color="gray",
        width=0.003,
        headwidth=8,
    )

    return ax



from typing import Callable, Optional

import torch.nn as nn

from rl4co.models.zoo.am import AttentionModelPolicy
from rl4co.models.zoo.ham.encoder import GraphHeterogeneousAttentionEncoder


class HeterogeneousAttentionModelPolicy(AttentionModelPolicy):
    """Heterogeneous Attention Model Policy based on https://ieeexplore.ieee.org/document/9352489.
    We re-declare the most important arguments here for convenience as in the paper.
    See :class:`rl4co.models.zoo.am.AttentionModelPolicy` for more details.

    Args:
        encoder: Encoder module. Can be passed by sub-classes
        env_name: Name of the environment used to initialize embeddings
        init_embedding: Model to use for the initial embedding. If None, use the default embedding for the environment
        embed_dim: Dimension of the embeddings
        num_encoder_layers: Number of layers in the encoder
        num_heads: Number of heads for the attention in encoder
        normalization: Normalization to use for the attention layers
        feedforward_hidden: Dimension of the hidden layer in the feedforward network
        sdpa_fn: Function to use for the scaled dot product attention
        **kwargs: keyword arguments passed to the :class:`rl4co.models.zoo.am.AttentionModelPolicy`
    """

    def __init__(
        self,
        encoder: nn.Module = None,
        env_name: str = "pdp",
        init_embedding: nn.Module = None,
        embed_dim: int = 128,
        num_encoder_layers: int = 3,
        num_heads: int = 8,
        normalization: str = "batch",
        feedforward_hidden: int = 512,
        sdpa_fn: Optional[Callable] = None,
        **kwargs,
    ):
        if encoder is None:
            encoder = GraphHeterogeneousAttentionEncoder(
                init_embedding=init_embedding,
                num_heads=num_heads,
                embed_dim=embed_dim,
                num_encoder_layers=num_encoder_layers,
                env_name=env_name,
                normalization=normalization,
                feedforward_hidden=feedforward_hidden,
                sdpa_fn=sdpa_fn,
            )
        else:
            encoder = encoder

        super(HeterogeneousAttentionModelPolicy, self).__init__(
            env_name=env_name,
            encoder=encoder,
            embed_dim=embed_dim,
            num_encoder_layers=num_encoder_layers,
            num_heads=num_heads,
            normalization=normalization,
            **kwargs,
        )



import math

from typing import Callable, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor

from rl4co.models.common import ImprovementEncoder
from rl4co.models.nn.attention import MultiHeadCompat
from rl4co.models.nn.ops import AdaptiveSequential, Normalization
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class Synth_Attention(nn.Module):
    def __init__(self, n_heads: int, input_dim: int) -> None:
        super().__init__()

        hidden_dim = input_dim // n_heads

        self.n_heads = n_heads
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim

        self.W_query = nn.Parameter(torch.Tensor(n_heads, input_dim, hidden_dim))
        self.W_key = nn.Parameter(torch.Tensor(n_heads, input_dim, hidden_dim))
        self.W_val = nn.Parameter(torch.Tensor(n_heads, input_dim, hidden_dim))

        self.score_aggr = nn.Sequential(
            nn.Linear(2 * n_heads, 2 * n_heads),
            nn.ReLU(inplace=True),
            nn.Linear(2 * n_heads, n_heads),
        )

        self.W_out = nn.Parameter(torch.Tensor(n_heads, hidden_dim, input_dim))

        self.init_parameters()

    # used for init nn.Parameter
    def init_parameters(self):
        for param in self.parameters():
            stdv = 1.0 / math.sqrt(param.size(-1))
            param.data.uniform_(-stdv, stdv)

    def forward(
        self, h_fea: torch.Tensor, aux_att_score: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        # h should be (batch_size, n_query, input_dim)
        batch_size, n_query, input_dim = h_fea.size()

        hflat = h_fea.contiguous().view(-1, input_dim)

        shp = (self.n_heads, batch_size, n_query, self.hidden_dim)

        # Calculate queries, (n_heads, batch_size, n_query, hidden_dim)
        Q = torch.matmul(hflat, self.W_query).view(shp)
        K = torch.matmul(hflat, self.W_key).view(shp)
        V = torch.matmul(hflat, self.W_val).view(shp)

        # Calculate compatibility (n_heads, batch_size, n_query, n_key)
        compatibility = torch.cat((torch.matmul(Q, K.transpose(2, 3)), aux_att_score), 0)

        attn_raw = compatibility.permute(
            1, 2, 3, 0
        )  # (batch_size, n_query, n_key, n_heads)
        attn = self.score_aggr(attn_raw).permute(
            3, 0, 1, 2
        )  # (n_heads, batch_size, n_query, n_key)
        heads = torch.matmul(
            F.softmax(attn, dim=-1), V
        )  # (n_heads, batch_size, n_query, hidden_dim)

        h_wave = torch.mm(
            heads.permute(1, 2, 0, 3)  # (batch_size, n_query, n_heads, hidden_dim)
            .contiguous()
            .view(
                -1, self.n_heads * self.hidden_dim
            ),  # (batch_size * n_query, n_heads * hidden_dim)
            self.W_out.view(-1, self.input_dim),  # (n_heads * hidden_dim, input_dim)
        ).view(batch_size, n_query, self.input_dim)

        return h_wave, aux_att_score


class SynthAttNormSubLayer(nn.Module):
    def __init__(self, n_heads: int, input_dim: int, normalization: str) -> None:
        super().__init__()

        self.SynthAtt = Synth_Attention(n_heads, input_dim)

        self.Norm = Normalization(input_dim, normalization)

    __call__: Callable[..., Tuple[torch.Tensor, torch.Tensor]]

    def forward(
        self, h_fea: torch.Tensor, aux_att_score: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        # Attention and Residual connection
        h_wave, aux_att_score = self.SynthAtt(h_fea, aux_att_score)

        # Normalization
        return self.Norm(h_wave + h_fea), aux_att_score


class FFNormSubLayer(nn.Module):
    def __init__(
        self, input_dim: int, feed_forward_hidden: int, normalization: str
    ) -> None:
        super().__init__()

        self.FF = (
            nn.Sequential(
                nn.Linear(input_dim, feed_forward_hidden, bias=False),
                nn.ReLU(inplace=True),
                nn.Linear(feed_forward_hidden, input_dim, bias=False),
            )
            if feed_forward_hidden > 0
            else nn.Linear(input_dim, input_dim, bias=False)
        )

        self.Norm = Normalization(input_dim, normalization)

    __call__: Callable[..., torch.Tensor]

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        # FF and Residual connection
        out = self.FF(input)
        # Normalization
        return self.Norm(out + input)


class N2SEncoderLayer(nn.Module):
    def __init__(
        self, n_heads: int, input_dim: int, feed_forward_hidden: int, normalization: str
    ) -> None:
        super().__init__()

        self.SynthAttNorm_sublayer = SynthAttNormSubLayer(
            n_heads, input_dim, normalization
        )

        self.FFNorm_sublayer = FFNormSubLayer(
            input_dim, feed_forward_hidden, normalization
        )

    __call__: Callable[..., Tuple[torch.Tensor, torch.Tensor]]

    def forward(
        self, h_fea: torch.Tensor, aux_att_score: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        h_wave, aux_att_score = self.SynthAttNorm_sublayer(h_fea, aux_att_score)
        return self.FFNorm_sublayer(h_wave), aux_att_score


class N2SEncoder(ImprovementEncoder):
    """Neural Neighborhood Search Encoder as in Ma et al. (2022)
    First embed the input and then process it with a Graph AttepdN2ntion Network.

    Args:
        embed_dim: Dimension of the embedding space
        init_embedding: Module to use for the initialization of the node embeddings
        pos_embedding: Module to use for the initialization of the positional embeddings
        env_name: Name of the environment used to initialize embeddings
        pos_type: Name of the used positional encoding method (CPE or APE)
        num_heads: Number of heads in the attention layers
        num_layers: Number of layers in the attention network
        normalization: Normalization type in the attention layers
        feedforward_hidden: Hidden dimension in the feedforward layers
    """

    def __init__(
        self,
        embed_dim: int = 128,
        init_embedding: nn.Module = None,
        pos_embedding: nn.Module = None,
        env_name: str = "pdp_ruin_repair",
        pos_type: str = "CPE",
        num_heads: int = 4,
        num_layers: int = 3,
        normalization: str = "layer",
        feedforward_hidden: int = 128,
    ):
        super(N2SEncoder, self).__init__(
            embed_dim=embed_dim,
            init_embedding=init_embedding,
            pos_embedding=pos_embedding,
            env_name=env_name,
            pos_type=pos_type,
            num_heads=num_heads,
            num_layers=num_layers,
            normalization=normalization,
            feedforward_hidden=feedforward_hidden,
        )

        self.pos_net = MultiHeadCompat(num_heads, embed_dim, feedforward_hidden)

        self.net = AdaptiveSequential(
            *(
                N2SEncoderLayer(
                    num_heads,
                    embed_dim,
                    feedforward_hidden,
                    normalization,
                )
                for _ in range(num_layers)
            )
        )

    def _encoder_forward(self, init_h: Tensor, init_p: Tensor) -> Tuple[Tensor, Tensor]:
        embed_p = self.pos_net(init_p)
        final_h, final_p = self.net(init_h, embed_p)

        return final_h, final_p
